'''Generate diagnostic plot on console environment
'''

import argparse
import sys
import numpy as np
from os.path import join, isfile, exists
import pkg_resources

from halla.utils.filesystem import reset_dir
from halla.utils.report import generate_lattice_plot
from .loader import HAllAPartialLoader

def parse_argument(args):
    parser = argparse.ArgumentParser(
        description='HAllA diagnostic plot generator')

    # --load parameters--
    parser.add_argument(
        '-i', '--input',
        help='Path to the output directory',
        required=True)
    parser.add_argument(
        '-n', '--block_num',
        help='Number of top clusters for generating the lattice plots; if -1, show all clusters',
        default=50, type=int, required=False)
    parser.add_argument(
        '--axis_stretch',
        help='Adding gaps to both ends of axis (for continuous data)',
        default=1e-5, type=float, required=False)
    parser.add_argument(
        '--plot_size',
        help='The size of each plot',
        default=4, type=float, required=False)
    parser.add_argument(
        '--file_type',
        help="The file type of the plots",
        default="pdf",
        required=False
    )
    parser.add_argument(
        '-o', '--output_dir',
        help='Directory name to store all the lattice plots under the HAllA/AllA result directory',
        default='diagnostic', required=False)
    parser.add_argument(
        '--dont_skip_large_blocks',
        required=False,
        dest='dont_skip',
        default=False,
        action='store_true')
    parser.add_argument(
        '--large_diagnostic_subset',
        help = "Subset the feature pairs plotted in large block (>15, <45) diagnostic plots.",
        required=False,
        dest='large_diagnostic_subset',
        default=105
    )
    return(parser.parse_args())

def main():
    params = parse_argument(sys.argv)
    input_dir = params.input
    loader = HAllAPartialLoader(input_dir)

    reset_dir(join(input_dir, params.output_dir))
    block_num = params.block_num
    if block_num == -1:
        block_num = len(loader.significant_blocks)
    else:
        block_num = min(block_num, len(loader.significant_blocks))
    for i, block in enumerate(loader.significant_blocks[:block_num]):
        title = 'Association %d' % (i+1)
        out_file = join(input_dir, params.output_dir, 'association_%d.' % (i+1) + params.file_type)
        warn_file = join(input_dir, params.output_dir, 'warnings.txt')
        x_data = loader.X.to_numpy()[block[0],:]
        y_data = loader.Y.to_numpy()[block[1],:]
        x_ori_data = loader.X_ori.to_numpy()[block[0],:]
        y_ori_data = loader.Y_ori.to_numpy()[block[1],:]
        x_features = loader.X_features[block[0]]
        y_features = loader.Y_features[block[1]]
        x_types = np.array(loader.X_types)[block[0]]
        y_types = np.array(loader.Y_types)[block[1]]
        if (x_data.shape[0] + y_data.shape[0]) > 15 and (x_data.shape[0] + y_data.shape[0]) <= 45:
            warn_string = "Over 15 features included in association %d. Only a subset of features will be shown in the diagnostic plot. Increase --large_diagnostic_subset beyond 105 to show more." % (i+1)
            if exists(warn_file):
                append_write = 'a'
            else:
                append_write = 'w'
            warn_file_write = open(warn_file, append_write)
            warn_file_write.write(warn_string + '\n')
            warn_file_write.close()
            print(warn_string, file = sys.stderr)
            generate_lattice_plot(x_data, y_data, x_ori_data, y_ori_data,
                                x_features, y_features, x_types, y_types, title,
                                out_file, axis_stretch=params.axis_stretch, plot_size=params.plot_size, n_pairs_to_show = params.large_diagnostic_subset)
            continue
        if (x_data.shape[0] + y_data.shape[0]) > 45 and not params.dont_skip:
            warn_string = "Skipping association %d because there are too many included features. Add --dont_skip_large_blocks to disable this behavior." % (i+1)
            if exists(warn_file):
                append_write = 'a'
            else:
                append_write = 'w'
            warn_file_write = open(warn_file, append_write)
            warn_file_write.write(warn_string + '\n')
            warn_file_write.close()
            print(warn_string, file = sys.stderr)
            continue
        generate_lattice_plot(x_data, y_data, x_ori_data, y_ori_data,
                                x_features, y_features, x_types, y_types, title,
                                out_file, axis_stretch=params.axis_stretch, plot_size=params.plot_size, n_pairs_to_show = (x_data.shape[0] + y_data.shape[0])**2)

if __name__ == "__main__":
    main()


import numpy as np

def compare_numpy_array(ar1, ar2):
    '''Compare and check if the values of two numpy arrays are equal
    '''
    all_float = False
    try:
        ar1 = ar1.astype(float)
        ar2 = ar2.astype(float)
        all_float = True
        np.testing.assert_allclose(ar1, ar2, verbose=True)
        return(True)
    except:
        if all_float: return(False)
        try:
            np.testing.assert_array_equal(ar1, ar2)
            return(True)
        except:
            return(False)


from .utils.similarity import get_similarity_function, similarity2distance

import scipy.cluster.hierarchy as sch
import scipy.spatial.distance as spd
import numpy as np

class HierarchicalTree(object):
    def __init__(self, matrix, pdist_metric, linkage_method, sim2dist_set_abs=True, sim2dist_func=None):
        '''Args:
        - matrix        : a pandas DataFrame object or a mxn array
        - pdist_metric  : the pairwise distance metric
        - linkage_method: the hierarchical linkage method
        '''
        if pdist_metric == 'xicor':
            treemetric = get_similarity_function('symmetric_xicor')
        else:
            treemetric = get_similarity_function(pdist_metric)

        self.distance_matrix = similarity2distance(spd.pdist(matrix, metric=treemetric),
                                                   sim2dist_set_abs,
                                                   sim2dist_func)
        self.distance_matrix = np.clip(self.distance_matrix, a_min=0, a_max=None)
        self.distance_matrix_sqr = spd.squareform(self.distance_matrix)
        self._generate_hierarchical_clusters(linkage_method)

    '''Private functions
    '''
    def _generate_hierarchical_clusters(self, linkage_method):
        # perform hierarchical clustering
        self.linkage = sch.linkage(self.distance_matrix, method=linkage_method)
        self.tree = sch.to_tree(self.linkage)



import pandas as pd
from os.path import join, isfile
import numpy as np
import scipy.cluster.hierarchy as sch
from re import search

from halla.utils.data import eval_type

class HAllAPartialLoader(object):
    def __init__(self, input_dir):
        self.input_dir = input_dir
        self.load_datasets()
        self.load_similarity_table()
        self.load_significant_clusters()
        self.load_linkages()
    
    def load_datasets(self):
        def fix_type(df, types):
            updated_df = df.copy(deep=True)
            for row_i in range(updated_df.shape[0]):
                updated_df.iloc[row_i] = updated_df.iloc[row_i].to_numpy().astype(types[row_i])
            return(updated_df)
        self.X_ori, self.X_types = eval_type(pd.read_table(join(self.input_dir, 'X_original.tsv'), index_col=0))
        self.Y_ori, self.Y_types = eval_type(pd.read_table(join(self.input_dir, 'Y_original.tsv'), index_col=0))
        self.X = fix_type(pd.read_table(join(self.input_dir, 'X.tsv'), index_col=0), self.X_types)
        self.Y = fix_type(pd.read_table(join(self.input_dir, 'Y.tsv'), index_col=0), self.Y_types)
        self.X_feat_map = { name: i for i, name in enumerate(self.X.index.to_list()) }
        self.Y_feat_map = { name: i for i, name in enumerate(self.Y.index.to_list()) }
        self.X_features = self.X.index.to_numpy()
        self.Y_features = self.Y.index.to_numpy()
        
    def load_similarity_table(self):
        df = pd.read_table(join(self.input_dir, 'all_associations.txt'))
        with open(join(self.input_dir, 'performance.txt')) as f:
            perf_lines = f.readlines()
        self.sim_table = np.zeros((self.X.shape[0], self.Y.shape[0]))
        perf_match = [search("fdr alpha", line) for line in perf_lines]
        perf_i = [i for i,v in enumerate(perf_match) if v != None]

        fdr_thresh = float(str.split(perf_lines[perf_i[0]], ' ')[-1][:-1]) # TODO make it find the line that says fdr alpha
        self.fdr_reject_table = np.zeros((self.X.shape[0], self.Y.shape[0]))
        for row in df.to_numpy():
            x, y = self.X_feat_map[row[0]], self.Y_feat_map[row[1]]
            self.sim_table[x][y] = row[2]
            self.fdr_reject_table[x][y] = row[4] < fdr_thresh

    def load_significant_clusters(self):
        df = pd.read_table(join(self.input_dir, 'sig_clusters.txt'))
        self.significant_blocks = []
        for row in df.to_numpy():
            X_feats, Y_feats = row[1].split(';'), row[2].split(';')
            block_0 = [self.X_feat_map[feat] for feat in X_feats]
            block_1 = [self.Y_feat_map[feat] for feat in Y_feats]
            block = [block_0, block_1]
            self.significant_blocks.append(block)
    
    def load_linkages(self):
        if not isfile(join(self.input_dir, 'X_linkage.npy')):
            self.name = 'AllA'
            return
        self.name = 'HAllA'
        self.X_linkage = np.load(join(self.input_dir, 'X_linkage.npy'))
        self.Y_linkage = np.load(join(self.input_dir, 'Y_linkage.npy'))
        self.X_tree = sch.to_tree(self.X_linkage)
        self.Y_tree = sch.to_tree(self.Y_linkage)


'''Run halla on console environment
'''

import argparse
import sys
import numpy as np
from os.path import join, basename, splitext, isdir
import pkg_resources

from halla import HAllA, AllA

def parse_argument(args):
    parser = argparse.ArgumentParser(
        description='HAllA: Hierarchical All-against-All significance association testing version ' + pkg_resources.require('HAllA')[0].version,
    )
    # --load parameters--
    parser.add_argument(
        '-x', '--x_file',
        help='Path to X dataset: a tab-delimited input file, one row per feature, one column per measurement',
        required=True)
    parser.add_argument(
        '-y', '--y_file',
        help='Path to Y dataset: a tab-delimited input file, one row per feature, one column per measurement',
        required=False)

    # --halla parameters--
    parser.add_argument(
        '--alla',
        help='Use AllA instead of HAllA',
        action='store_true', required=False)
    parser.add_argument(
        '--max_freq_thresh',
        help='The maximum frequency threshold - features with max frequences >= the threshold will be removed',
        default=1, type=float, required=False)
    parser.add_argument(
        '--transform_data_funcs',
        help='Continuous data transformation function - a list',
        nargs='+',
        default=None, required=False)
    parser.add_argument(
        '--disable_bypass_discretization_if_possible',
        help='Disable bypassing discretization when all features are continuous',
        action='store_true', required=False)
    parser.add_argument(
        '--discretize_func',
        help='Discretization - function {None, quantile, kmeans, uniform, jenks}',
        default=None, choices=['None', 'quantile', 'kmeans', 'uniform', 'jenks'], required=False)
    parser.add_argument(
        '--discretize_num_bins',
        help='Discretization - number of bins',
        default=None, type=int, required=False)
    parser.add_argument(
        '-m', '--pdist_metric',
        help='Distance/similarity metric {spearman, pearson, dcor, mi, nmi, xicor}',
        default='spearman', choices=['spearman', 'pearson', 'dcor', 'mi', 'nmi', 'xicor'], required=False)
    parser.add_argument(
        '--sim2dist_disable_abs',
        help='Hierarchical clustering - disable setting similarity scores as absolute when computing distance',
        action='store_true', required=False)
    parser.add_argument(
        '--linkage_method',
        help='Hierarchical clustering linkage method - check scipy.cluster.hierarchy.linkage()',
        default='average', required=False)
    parser.add_argument(
        '--permute_func',
        help='P-value approximation function in the p-value permutation test',
        default='gpd', choices=['gpd', 'ecdf'], required=False)
    parser.add_argument(
        '--permute_iters',
        help='# iterations in the p-value permutation test',
        default=1000, type=int, required=False)
    parser.add_argument(
        '--disable_permute_speedup',
        help='Disable breaking early in the permutation test if p-value is insignificant',
        action='store_true', required=False)
    parser.add_argument(
        '--fdr_alpha',
        help='FDR threshold',
        default=0.05, type=float, required=False)
    parser.add_argument(
        '--fdr_method',
        help='FDR method - check statsmodels.stats.multitest.multipletests()',
        default='fdr_bh', required=False)
    parser.add_argument(
        '--fnr_thresh',
        help='FNR threshold',
        default=0.2, type=float, required=False)
    parser.add_argument(
        '--rank_cluster',
        help='Procedure to rank cluster using the p-values within the cluster {best, average}',
        default='best', choices=['best', 'average'], required=False)
    parser.add_argument(
        '-o', '--out_dir',
        help='Directory path to store results', required=True)
    parser.add_argument(
        '--disable_verbose',
        help='Disable verbose',
        action='store_true', required=False)
    parser.add_argument(
        '--seed',
        help='Randomization seed',
        default=None, type=float, required=False)

    # --hallagram parameters--
    parser.add_argument(
        '--hallagram',
        help='Generates hallagram',
        default=True,action='store_true', required=False)
    parser.add_argument(
        '--no_hallagram',
        help='Turn off the automatically generated hallagram',
        dest='hallagram',
        action='store_false')

    # --clustermap parameters--
    parser.add_argument(
        '--clustermap',
        help='Generates clustermap',
        action='store_true', required=False)

    # --hallagram/clustermap parameters--
    parser.add_argument(
        '--x_dataset_label',
        help='Hallagram/clustermap: label for X dataset',
        default='', required=False)
    parser.add_argument(
        '--y_dataset_label',
        help='Hallagram/clustermap: label for Y dataset',
        default='', required=False)
    parser.add_argument(
        '--cbar_label',
        help='Hallagram/clustermap: label for the colorbar',
        default='', required=False)
    parser.add_argument(
        '-n', '--block_num',
        help='Number of top clusters to show (for hallagram only); if -1, show all clusters',
        default=50, type=int, required=False)
    parser.add_argument(
        '--trim',
        help='Trim hallagram to features containing at least one significant block',
        default=True,
        type=bool, required=False)
    parser.add_argument(
        "--plot_file_type",
        help = "File type of hallagram output",
        default = "pdf",
        dest="plot_type",
        required = False
    )

    # --other options--
    parser.add_argument(
        '--no_progress',
        help="Turn off the progress bar for p-value table calculations",
        dest='no_progress',
        default=False,
        action = 'store_true',required=False)
    parser.add_argument(
        '--dont_copy',
        help="Don't make a copy of the data files in the output directory",
        dest='dont_copy',
        default=False,
        action = 'store_true',required=False)
    parser.add_argument(
        '--force_permutations',
        help="If turned on, force permutation testing",
        dest='force_permutations',
        default=False,
        action = 'store_true',required=False)
    parser.add_argument(
        '--num_threads',
        help='Number of threads to use when running permutation tests in parallel, default=4',
        default=1, type=int, required=False)
    parser.add_argument(
        '--version',
        action='version',
        version=pkg_resources.require('HAllA')[0].version
    )
    parser.add_argument(
        '--dont_skip_large_blocks',
        required=False,
        help="Don't skip very large (>45 features) blocks in diagnostic plots",
        dest='dont_skip',
        default=False,
        action='store_true')
    parser.add_argument(
        '--large_diagnostic_subset',
        help = "Subset the feature pairs plotted in large block (>15, <45) diagnostic plots.",
        required=False,
        dest='large_diagnostic_subset',
        default=105
    )
    parser.add_argument(
        '--splitting_diagnostic_mode',
        required=False,
        dest='splitting_diagnostic_mode',
        help="Diagnostic mode to write out tree descent algorithm progress. Prints branches being considered at each step and Gini score improvement (the latter only if applicable). Values listed in brackets are indices (0-indexed) of features in X and Y datasets respectively.",
        default=False,
        action='store_true')
    parser.add_argument(
        '--gini_uncertainty_level',
        required=False,
        dest='gini_uncertainty_level',
        type=float,
        help="Gini uncertainty mode opts to split larger hierarchical branches when the difference in Gini impurity improvement is less than the given level.",
        default=0.02)

    # --diagnostic-plot parameters--
    parser.add_argument(
        '--diagnostic_plot',
        help='Generates diagnostic plot',
        action='store_true', required=False)
    params = parser.parse_args()
    if params.discretize_func == 'None':
        params.discretize_func = None
    return(parser.parse_args())

def main():
    params = parse_argument(sys.argv)

    if isdir(params.out_dir):
        sys.exit("Error: output directory already exists.")

    if params.alla:
        instance = AllA(max_freq_thresh=params.max_freq_thresh,
                 transform_data_funcs=params.transform_data_funcs,
                 discretize_bypass_if_possible=not params.disable_bypass_discretization_if_possible,
                 discretize_func=params.discretize_func, discretize_num_bins=params.discretize_num_bins,
                 pdist_metric=params.pdist_metric,
                 permute_func=params.permute_func, permute_iters=params.permute_iters,
                 permute_speedup=not params.disable_permute_speedup,
                 fdr_alpha=params.fdr_alpha, fdr_method=params.fdr_method,
                 out_dir=params.out_dir, verbose=not params.disable_verbose,
                 no_progress=params.no_progress, dont_copy=params.dont_copy, force_permutations=params.force_permutations, 
                 num_threads=params.num_threads, 
                 splitting_diagnostic_mode=params.splitting_diagnostic_mode,
                 gini_uncertainty_level=params.gini_uncertainty_level,
                 seed=params.seed)
    else:
        instance = HAllA(max_freq_thresh=params.max_freq_thresh,
                 transform_data_funcs=params.transform_data_funcs,
                 discretize_bypass_if_possible=not params.disable_bypass_discretization_if_possible,
                 discretize_func=params.discretize_func, discretize_num_bins=params.discretize_num_bins,
                 pdist_metric=params.pdist_metric, linkage_method=params.linkage_method,
                 sim2dist_set_abs=not params.sim2dist_disable_abs,
                 permute_func=params.permute_func, permute_iters=params.permute_iters,
                 permute_speedup=not params.disable_permute_speedup,
                 fdr_alpha=params.fdr_alpha, fdr_method=params.fdr_method,
                 fnr_thresh=params.fnr_thresh, rank_cluster=params.rank_cluster,
                 out_dir=params.out_dir, verbose=not params.disable_verbose,
                 no_progress=params.no_progress, dont_copy=params.dont_copy, force_permutations=params.force_permutations,
                 dont_skip=params.dont_skip,
                 num_threads=params.num_threads,
                 splitting_diagnostic_mode=params.splitting_diagnostic_mode,
                 gini_uncertainty_level=params.gini_uncertainty_level,
                 seed=params.seed)
    instance.load(params.x_file, params.y_file)
    instance.run()
    if params.clustermap:
        if params.x_dataset_label=='':
            params.x_dataset_label = splitext(basename(params.x_file))[0]
        if params.y_dataset_label=='':
            params.y_dataset_label = splitext(basename(params.y_file))[0]
        if params.alla:
            print('AllA does not produce clustermap.', file = sys.stderr)
        else:
            instance.generate_clustermap(x_dataset_label=params.x_dataset_label,
                                           y_dataset_label=params.y_dataset_label,
                                           cbar_label=params.cbar_label)
    if params.hallagram:
        if params.x_dataset_label=='':
            params.x_dataset_label = splitext(basename(params.x_file))[0]
        if params.y_dataset_label=='':
            params.y_dataset_label = splitext(basename(params.y_file))[0]
        instance.generate_hallagram(x_dataset_label=params.x_dataset_label,
                                           y_dataset_label=params.y_dataset_label,
                                           cbar_label=params.cbar_label, trim = params.trim,
                                           plot_type = params.plot_type)
    if params.diagnostic_plot:
        if params.alla:
            print('AllA does not produce diagnostic plot.', file = sys.stderr)
        else:
            instance.generate_diagnostic_plot()

if __name__ == "__main__":
    main()



'''Generate synthetic dataset X and Y

Generates: a directory that contains 3 files:
- X, Y: a paired dataset
- A   : contains the associated features between the paired dataset
'''

import argparse
import sys
import numpy as np
import itertools
import pandas as pd
from os.path import join
from scipy.stats import ortho_group
import math

from halla.utils.data import discretize_vector
from halla.utils.filesystem import reset_dir

def parse_argument(args):
    parser = argparse.ArgumentParser(
        description='HAllA synthetic data generator - produces a pair of datasets X & Y with specified association among their features'
    )
    parser.add_argument('-n', '--samples', help='# samples in both X and Y', default=50, type=int, required=False)
    parser.add_argument('-xf', '--xfeatures', help='# features in X', default=500, type=int, required=False)
    parser.add_argument('-yf', '--yfeatures', help='# features in Y', default=500, type=int, required=False)
    parser.add_argument('-b', '--blocks', help='# significant blocks; default = min(xfeatures, yfeatures, samples)/3',
                        default=None, type=int, required=False)
    parser.add_argument('-a', '--association', help='association type {line, parabola, log, sine, step, mixed, categorical}; default: line',
                        default='line', choices=['line', 'parabola', 'log', 'sine', 'step', 'mixed', 'categorical'], required=False)
    parser.add_argument('-nw', '--noise-within', dest='noise_within', help='noise within blocks [0 (no noise)..1 (complete noise)]',
                        default=0.25, type=float, required=False)
    parser.add_argument('-nb', '--noise-between', dest='noise_between', help='noise between associated blocks [0 (no noise)..1 (complete noise)]',
                        default=0.25, type=float, required=False)
    parser.add_argument('-o', '--output', help='the output directory', required=True)

    # check requirements
    params = parser.parse_args()
    # samples must be > 0
    if params.samples <= 0: raise ValueError('# samples must be > 0')
    # xfeatures and yfeatures must be > 0
    if params.xfeatures <= 0 or params.yfeatures <= 0: raise ValueError('# features must be > 0')
    # blocks must be 1 .. min(5, min(xfeatures, yfeatures)/2)
    if params.blocks is None:
        params.blocks = min(params.xfeatures, params.yfeatures, params.samples)//3
    if not (params.blocks > 0 and params.blocks <= min(params.xfeatures, params.yfeatures, params.samples)/3):
        raise ValueError('# blocks is invalid; must be [1..min(xfeatures, yfeatures, samples)/3]')
    # noises must be [0..1]
    if params.noise_between < 0 or params.noise_between > 1 or \
        params.noise_within < 0 or params.noise_within > 1:
        raise ValueError('Noise within/between must be [0..1]')
    return(params)

def run_data_generator(sample_num=50, features_num=(500, 500), block_num=5, association='line',
                        noise_within=0.25, noise_between=0.25, noise_within_std=0.7, noise_between_std=0.7):
    '''Generate synthetic data with the following steps:
    1) generate a base B [-1, 1] from uniform distribution
    2) derive base_X and base_Y from B with noise = between_noise
    3) derive features in X and Y from base_X and base_Y with noise = within_noise

    Available associations include:
    - line       : X = base_X + noise; Y = base_Y + noise
    - parabola   : X = base_X + noise; Y = base_Y * base_Y + noise
    - log        : base = abs(base); X = base_X + noise; Y = log(abs(base_Y)) + noise
    - sine       : base = base * 2 ; X = base_X + noise; Y = 2 * sin(pi * base_Y) + noise
    - step       : X = base_X + noise; Y = { div base_Y into 4 quantiles; p1 = 2.0, p2 = 1.0, p3 = 3.0, p4 = 0.0 } + noise
    - categorical: same as step; discretize all features into {[3..6]} bins
    - mixed      : same as step; discretize some features in X and Y into into {[3..6]} bins
    '''
    def create_base():
        '''Generate base matrix [block_num x sample_num] with rows independent to each other
        '''
        # generate orthogonal matrix from uniform distribution
        #   then pick block_num rows given block_num <= sample_num
        base = ortho_group.rvs(sample_num)[:block_num]
        # enlarge the range to around [-1, 1]
        mult = min(abs(1.0 / base.min()), abs(1.0 / base.max()))
        base = base * mult
        # test if rows are independent
        test = base @ base.T
        np.testing.assert_allclose(test, mult*mult*np.eye(block_num), atol=1e-10, err_msg='The rows in base are not orthonormal')
        return(base)

    def div_features_into_blocks(feat_num):
        # initialize
        blocks_size = [0] * block_num
        assoc = [[]] * block_num
        # obtain block size
        for _ in range(feat_num): blocks_size[np.random.choice(block_num)] += 1
        # assign feature indices to blocks
        start_idx = 0
        for i in range(block_num):
            assoc[i] = [i for i in range(start_idx, start_idx + blocks_size[i])]
            start_idx = start_idx + blocks_size[i]
        return(assoc)

    def abs_if_necessary(a):
        if association == 'log': return(np.abs(a))
        return(a)

    # initialize all matrices
    x_feat_num, y_feat_num = features_num
    X, Y = np.zeros((x_feat_num, sample_num)), np.zeros((y_feat_num, sample_num))
    A = np.zeros(features_num)

    # step 1: generate base
    base = abs_if_necessary(create_base())
    if association == 'sine': base = 2 * base # for spreading out x

    # assign features in X and Y to blocks
    x_assoc, y_assoc = div_features_into_blocks(x_feat_num), div_features_into_blocks(y_feat_num)
    for block_i in range(block_num):
        # step 2.1: derive base_X from base given noise_between
        base_X = base[block_i] + noise_between * np.random.normal(scale=noise_between_std, size=1)
        # step 3.1: derive X from base_X given noise_within
        for feat_x in x_assoc[block_i]:
            X[feat_x] = base_X + noise_within * np.random.normal(scale=noise_within_std, size=sample_num)

        # determine positive or negative association if appropriate; arbitrary probs
        sign_corr = np.random.choice([-1, 1], p=[0.4, 0.6])

        # step 2.2: derive base_Y from base given noise_between
        base_Y = abs_if_necessary(base[block_i] + noise_between * np.random.normal(scale=noise_between_std, size=1))
        # step 3.2: derive Y from base_Y given noise_within
        for feat_y in y_assoc[block_i]:
            if association == 'line':
                Y[feat_y] = sign_corr * base_Y
            elif association == 'parabola':
                Y[feat_y] = sign_corr * base_Y * base_Y
            elif association == 'log':
                Y[feat_y] = np.log(base_Y)
            elif association == 'sine':
                Y[feat_y] = 2 * np.sin(math.pi * base_Y)
            elif association == 'step':
                # divide base_Y into 4 quantiles
                p1, p2, p3 = np.percentile(base_Y, 25), np.percentile(base_Y, 50), np.percentile(base_Y, 75)
                Y[feat_y] = [2.0 if val < p1 else \
                             1.0 if val < p2 else \
                             3.0 if val < p3 else 0.0 for val in base_Y]
            else:
                # default for now
                Y[feat_y] = sign_corr * base_Y
            Y[feat_y] = Y[feat_y] + noise_within * np.random.normal(scale=noise_within_std, size=sample_num)
        # update A
        for i, j in itertools.product(x_assoc[block_i], y_assoc[block_i]):
            A[i][j] = 1
    if association in ['categorical', 'mixed']:
        X_new = np.empty((x_feat_num, sample_num), dtype=object)
        Y_new = np.empty((y_feat_num, sample_num), dtype=object)
        # select features to be discretized
        if association == 'categorical':
            xdisc_feat_indices = [i for i in range(x_feat_num)]
            ydisc_feat_indices = [j for j in range(y_feat_num)]
        else:
            xdisc_feat_num, ydisc_feat_num = np.random.choice(x_feat_num), np.random.choice(y_feat_num)
            xdisc_feat_indices = np.random.choice(x_feat_num, xdisc_feat_num, replace=False)
            ydisc_feat_indices = np.random.choice(y_feat_num, ydisc_feat_num, replace=False)
        for i in range(x_feat_num):
            if i in xdisc_feat_indices:
                discretized = discretize_vector(X[i], func='quantile',
                                num_bins=min(np.random.choice(range(3,7)), sample_num//2))
                X_new[i] = [chr(int(val) + 65) for val in discretized]
            else:
                X_new[i] = X[i]
        for j in range(y_feat_num):
            if j in ydisc_feat_indices:
                discretized = discretize_vector(Y[j], func='quantile',
                                num_bins=min(np.random.choice(range(3,7)), sample_num//2))
                Y_new[j] = [chr(int(val) + 65) for val in discretized]
            else:
                Y_new[j] = Y[j]
        return(X_new, Y_new, A)
    return(X, Y, A)

def store_tables(X, Y, A, association, out_dir):
    '''Store generated tables X,Y,A into files under out_dir directory
    '''

    def create_df(table, col_pref, row_pref):
        '''Create pandas DataFrame from table given:
        - table   : a 2D numpy array
        - col_pref: the column prefix
        - row_pref: the row prefix
        '''
        return pd.DataFrame(
            data={ '%s%d' % (col_pref, j): table[:,j] for j in range(table.shape[1]) },
            index=['%s%d' % (row_pref, i) for i in range(table.shape[0])]
        )

    x_feat_num, sample_num = X.shape
    y_feat_num, _ = Y.shape

    # create directory
    reset_dir(out_dir)

    # store df in files
    filename_format = '%s_%s_%s_%s.txt' % ('%s', association, '%s', '%d')
    dataset_format = filename_format % ('%s', '%d', sample_num)
    create_df(X, 'S', 'X').to_csv(join(out_dir, dataset_format % ('X', x_feat_num)), sep='\t', index=True)
    create_df(Y, 'S', 'Y').to_csv(join(out_dir, dataset_format % ('Y', y_feat_num)), sep='\t', index=True)
    create_df(A, 'Y', 'X').to_csv(join(out_dir, filename_format % ('A', x_feat_num, y_feat_num)), sep='\t', index=True)

def main():
    # parse arguments
    params = parse_argument(sys.argv)
    # generate datasets
    X, Y, A = run_data_generator(params.samples, (params.xfeatures, params.yfeatures), params.blocks,
                                    params.association, params.noise_within, params.noise_between)
    # store datasets
    store_tables(X, Y, A, params.association, out_dir=params.output)

if __name__ == "__main__":
    main()



import numpy as np
import math
from scipy.stats import rankdata, entropy, zscore, rankdata
from sklearn.preprocessing import KBinsDiscretizer, quantile_transform
import pandas as pd
import jenkspy
import warnings

def eval_type(df):
    '''Evaluate and set the type for each feature given dataframe df
       where each row represents one feature
    Return a tuple (updated_df, all_cont):
    - updated_df: df with updated type
    - types     : type for each row in np array
    '''
    def _get_type(np_array):
        '''Assess if the numpy array type is continuous
        #TODO: allow users to manually specify types, esp categorical integers
        '''
        try:
            np_array.astype(float)
            return(float)
        except:
            return(object)

    if not isinstance(df, pd.DataFrame):
        raise ValueError('The argument for eval_type() should be a pandas DataFrame!')
    # update all NaNs to None
    updated_df = df.copy(deep=True)
    types = []
    for row_i in range(updated_df.shape[0]):
        row_type = _get_type(updated_df.iloc[row_i].to_numpy())
        updated_df.iloc[row_i] = updated_df.iloc[row_i].to_numpy().astype(row_type)
        types.append(row_type)
    return(updated_df, np.array(types))

def discretize_vector(ar, ar_type=float, func=None, num_bins=None):
    '''Discretize vector to [0 .. (bin_num - 1)], given:
    - ar      : 1D numpy vector
    - ar_type : the type of the vector
    - func    : discretization function
    - num_bins: # bins
    '''
    def _discretize_categorical(ar):
        '''Represent the categories in integer
        '''
        new_ar = []
        flag, counter = {}, 0
        for item in ar:
            if item != item: item = 'nan'
            if item not in flag:
                flag[item] = counter
                counter += 1
            new_ar.append(flag[item])
        return(np.array(new_ar))

    def _discretize_continuous(ar, func=None, num_bins=None):
        '''Assign continuous vector to bins;
        available functions:
        - quantile, uniform, kmeans (https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.KBinsDiscretizer.html)
        '''
        if func is None: # no discretization is needed
            return(ar)
        if num_bins is None:
            # by default, num_bins = cube_root(n) or # unique(n)
            num_bins = min(math.floor(len(ar)**(1./3.)), len(set(ar)))
        elif num_bins == 0:
            raise ValueError('# bins should be > 0')
        else:
            num_bins = min(num_bins, len(set(ar)))
        if func in ['quantile', 'uniform', 'kmeans']:
            warnings.simplefilter('ignore') # ignore nan warning
            # remove NaNs temporarily since KBinsDiscretizer currently doesn't handle NaNs
            temp_ar = np.array([x for x in ar if x == x])
            nans = np.array([not x == x for x in ar])
            discretizer = KBinsDiscretizer(n_bins=num_bins, encode='ordinal', strategy=func)
            temp_discretized_result = discretizer.fit_transform(temp_ar.reshape(len(temp_ar), 1)).reshape(len(temp_ar))
            # assign NaNs to a separate bin
            discretized_result = np.zeros(len(ar))
            discretized_result[~nans] = temp_discretized_result
            discretized_result[nans] = np.nan
            warnings.resetwarnings()
        elif func == 'jenks':
            warnings.simplefilter('ignore') # ignore nan warning
            # remove the first lower bound
            breaks = jenkspy.jenks_breaks(ar, nb_class=num_bins)
            breaks[-1] += 1
            discretized_result = []
            for val in ar:
                if np.isnan(val):
                    discretized_result.append(np.nan)
                    continue
                for bound_i in range(1, len(breaks)):
                    if val >= breaks[bound_i-1] and val < breaks[bound_i]:
                        discretized_result.append(bound_i-1)
                        break
            discretized_result = np.array(discretized_result, dtype=float)
            warnings.resetwarnings()
        else:
            raise ValueError('Discretization function not available...')
        # tidy up data in case some bins are empty
        counter = 0
        for i in range(int(np.nanmax(discretized_result) + 1)):
            if i not in discretized_result: continue
            discretized_result[discretized_result == i] = counter
            counter += 1
        # assign missing values to a separate bin
        discretized_result[np.isnan(discretized_result)] = np.nanmax(discretized_result) + 1
        return(discretized_result)

    if ar_type == object:
        return(_discretize_categorical(ar))
    return(_discretize_continuous(ar, func, num_bins))

def keep_feature(x, max_freq_thresh=1):
    '''Decide if the feature should be kept based on the maximum frequency threshold, given:
    - x              : the vector from a DataFrame row
    - max_freq_thresh: the threshold for maximum frequency in fraction [0..1], disabled if None
    '''
    x = x.to_list()
    _, count = np.unique(x, return_counts=True)
    if max_freq_thresh is None: return(True)
    return(count.max() / count.sum() < max_freq_thresh)

def transform(df, types, funcs=None):
    '''Transform the continuous rows in the dataframe given function
    Available functions:
    - zscore, rank, quantile
    - any function available in numpy attributes, e.g., log, arcsin, sqrt
    '''
    warnings.filterwarnings('error') # to catch RuntimeError

    if funcs is None: return(df)

    if not isinstance(funcs, list): funcs = [funcs]
    funcs = [func.lower() for func in funcs]
    # ensure all functions are valid
    for func in funcs:
        if func not in ['zscore', 'rank', 'quantile'] and not hasattr(np, func):
            raise ValueError('The transform function should be an attribute of numpy or one of {zscore, rank, quantile}.')

    updated_df = df.copy(deep=True)
    for row_i, type_i in enumerate(types):
        if type_i == object: continue
        row = updated_df.iloc[row_i].to_numpy(dtype=float)
        for func in funcs:
            try:
                if func == 'zscore':
                    row = zscore(row, nan_policy='omit')
                elif func == 'rank':
                    row = rankdata(row)
                elif func == 'quantile':
                    row = quantile_transform(row.reshape((len(row), 1)), n_quantiles=len(row), output_distribution='normal', random_state=0).reshape(len(row))
                elif func == 'sqrt':
                    row = np.sqrt(np.abs(row)) * np.sign(row)
                else:
                    row = getattr(np, func)(row)
            except RuntimeWarning:
                raise ValueError('Runtime error in transforming data: invalid value encountered')
        updated_df.iloc[row_i] = row
    warnings.resetwarnings()
    return(updated_df)

def preprocess(df, types, transform_funcs=None, max_freq_thresh=0.8, discretize_func=None, discretize_num_bins=None):
    '''Preprocess input data
    1) remove features with low entropy
    2) transform if the function is specified
    3) discretize values if needed

    Args:
    - df                 : a panda dataframe
    - types              : a numpy list indicating the type of each feature
    - transform_funcs    : list of ordered functions to transform continuous data
    - max_freq_thresh    : the threshold for maximum frequency in fraction [0..1], disabled if None
    - discretize_func    : function for discretizing
    - discretize_num_bins: # bins for discretizing #TODO: different bins for different features?
    '''
    # 1) remove features with low entropy
    kept_rows = df.apply(keep_feature, 1, max_freq_thresh=max_freq_thresh)
    types, df = types[kept_rows], df[kept_rows]

    # 2) transform if the function is specified
    df = transform(df, types, funcs=transform_funcs)

    # 3) discretize values if needed
    updated_df = df.copy(deep=True)
    for row_i, type_i in enumerate(types):
        updated_df.iloc[row_i] = discretize_vector(updated_df.iloc[row_i].to_numpy().astype(type_i), type_i,
                                                   func=discretize_func,
                                                   num_bins=discretize_num_bins)
    # re-run the feature removal because sometimes discretization can reduce entropy:
    # https://github.com/scikit-learn/scikit-learn/issues/19433
    # ^ This sort of thing happens when there are sparse, small values. See case_studies/HMP data , 2nd row
    kept_rows = updated_df.apply(keep_feature, 1, max_freq_thresh=max_freq_thresh)
    types, updated_df = types[kept_rows], updated_df[kept_rows]

    return(updated_df, df, types)

def is_all_cont(types):
    '''Check if all types are continuous
    '''
    return(np.sum(types == object) == 0)



from .config_loader import config, update_config
from .hierarchy import HierarchicalTree
from .logger import HAllALogger
from .utils.data import preprocess, eval_type, is_all_cont
from .utils.similarity import get_similarity_function
from .utils.stats import get_pvalue_table, pvalues2qvalues, test_pvalue_run_time
from .utils.tree import compare_and_find_dense_block, trim_block
from .utils.report import generate_hallagram, generate_clustermap, \
                          report_all_associations, report_significant_clusters, \
                          generate_lattice_plot
from .utils.filesystem import reset_dir

import pandas as pd
import numpy as np
import scipy.spatial.distance as spd
from os.path import join, exists
from os import getcwd
import time
import sys

########
# AllA
########
class AllA(object):
    def __init__(self, max_freq_thresh=config.preprocess['max_freq_thresh'],
                 transform_data_funcs=config.preprocess['transform_funcs'],
                 discretize_bypass_if_possible=config.preprocess['discretize_bypass_if_possible'],
                 discretize_func=config.preprocess['discretize_func'], discretize_num_bins=config.preprocess['discretize_num_bins'],
                 pdist_metric=config.association['pdist_metric'],
                 permute_func=config.permute['func'], permute_iters=config.permute['iters'], permute_speedup=config.permute['speedup'],
                 fdr_alpha=config.stats['fdr_alpha'], fdr_method=config.stats['fdr_method'],
                 out_dir=config.output['dir'], verbose=config.output['verbose'], no_progress=False, dont_copy=False, force_permutations=False,
                 num_threads=4, dont_skip=False, large_diagnostic_subset=105, splitting_diagnostic_mode=False, gini_uncertainty_level = .02, seed=0):
        # update AllA config setting
        update_config('output', dir=out_dir, verbose=verbose)
        update_config('preprocess', max_freq_thresh=max_freq_thresh,
                                    transform_funcs=transform_data_funcs,
                                    discretize_bypass_if_possible=discretize_bypass_if_possible,
                                    discretize_func=discretize_func, discretize_num_bins=discretize_num_bins)
        update_config('association', pdist_metric=pdist_metric)
        update_config('permute', func=permute_func, iters=permute_iters, speedup=permute_speedup)
        update_config('stats', fdr_alpha=fdr_alpha, fdr_method=fdr_method)
        self._reset_attributes()
        self.no_progress = no_progress
        self.dont_copy = dont_copy
        self.force_permutations = force_permutations
        self.num_threads = num_threads
        self.dont_skip = dont_skip
        self.large_diagnostic_subset = large_diagnostic_subset
        self.verbose = verbose
        self.seed = seed
        self.splitting_diagnostic_mode = splitting_diagnostic_mode
        self.gini_uncertainty_level = gini_uncertainty_level

        if (out_dir == ".") or (out_dir == "./") or (out_dir == getcwd()):
            raise ValueError("Please specify an output directory other than the current directory.")

        if not hasattr(self, 'name'):
            self.name = 'AllA'
            self.logger = HAllALogger(name=self.name, config=config)

    '''Private functions
    '''
    def _reset_attributes(self):
        self.X, self.Y = None, None
        self.X_types, self.Y_types = None, None
        self.similarity_table = None
        self.pvalue_table, self.qvalue_table = None, None
        self.fdr_reject_table = None
        self.significant_blocks = None
        self.significant_blocks_qvalues = None
        self.has_loaded = False
        self.has_run = False
        self.verbose = False

    def _compute_pairwise_similarities(self):
        dist_metric = config.association['pdist_metric']

        self.logger.log_step_start('Step 1: Computing pairwise similarities, p-values, and q-values', sub=True)
        start_time = time.time()

        X, Y = self.X.to_numpy(), self.Y.to_numpy()

        # obtain similarity matrix
        self.logger.log_message('Generating the similarity table...')
        self.similarity_table = spd.cdist(X, Y, metric=get_similarity_function(dist_metric))

        # obtain p-values
        self.logger.log_message('Generating the p-value table...')
        confp = config.permute
        extrapolated_time, timing_message = test_pvalue_run_time(X, Y, pdist_metric=dist_metric,
                                                   permute_func=confp['func'], permute_iters=confp['iters'],
                                                   permute_speedup=confp['speedup'],
                                                   alpha=config.stats['fdr_alpha'],
                                                   force_perms = self.force_permutations,
                                                   num_threads=self.num_threads,
                                                   seed=self.seed)
        if extrapolated_time > 10 and self.verbose:
            self.logger.log_message(timing_message)

        self.pvalue_table = get_pvalue_table(X, Y, pdist_metric=dist_metric,
                                                   permute_func=confp['func'], permute_iters=confp['iters'],
                                                   permute_speedup=confp['speedup'],
                                                   alpha=config.stats['fdr_alpha'],
                                                   no_progress=self.no_progress,
                                                   force_permutations=self.force_permutations,
                                                   num_threads=self.num_threads,
                                                   seed=self.seed)

        # obtain q-values
        self.logger.log_message('Generating the q-value table...')
        self.fdr_reject_table, self.qvalue_table = pvalues2qvalues(self.pvalue_table.flatten(), config.stats['fdr_method'], config.stats['fdr_alpha'])
        self.qvalue_table = self.qvalue_table.reshape(self.pvalue_table.shape)
        self.fdr_reject_table = self.fdr_reject_table.reshape(self.pvalue_table.shape)

        end_time = time.time()
        self.logger.log_result('Number of significant associations', self.fdr_reject_table.sum())
        self.logger.log_step_end('Computing pairwise similarities, p-values, q-values', end_time - start_time, sub=True)

    def _find_dense_associated_blocks(self):
        '''Find significant cells based on FDR reject table
        '''
        def compare_qvalue(x):
            return(self.qvalue_table[x[0][0], x[1][0]])

        self.logger.log_step_start('Step 2: Finding densely associated blocks', sub=True)
        start_time = time.time()

        n, m = self.X.shape[0], self.Y.shape[0]
        self.significant_blocks = [[[x], [y]] for x in range(n) for y in range(m) if self.fdr_reject_table[x][y]]
        # sort by the p-values in ascending order
        self.significant_blocks.sort(key=compare_qvalue)
        self.significant_blocks_qvalues = [self.qvalue_table[x[0][0]][x[1][0]] for x in self.significant_blocks]

        end_time = time.time()
        self.logger.log_result('Number of significant clusters', len(self.significant_blocks))
        self.logger.log_step_end('Finding densely associated blocks', end_time - start_time, sub=True)

    def _generate_reports(self):
        '''Generate reports and store in config.output['dir'] directory:
        1) all_associations.txt: stores the associations between each feature in X and Y along with its
                                p-values and q-values in a table
        2) sig_clusters.txt    : stores only the significant clusters
        '''
        self.logger.log_step_start('Generating reports')

        # create directory
        dir_name = config.output['dir']
        reset_dir(dir_name, verbose=config.output['verbose'])

        # generate performance.txt
        self.logger.write_performance_log(dir_name, config)

        # generate all_associations.txt
        report_all_associations(dir_name,
                                self.X.index.to_numpy(),
                                self.Y.index.to_numpy(),
                                self.similarity_table,
                                self.pvalue_table,
                                self.qvalue_table)

        # generate sig_clusters.txt
        report_significant_clusters(dir_name,
                                    self.significant_blocks,
                                    self.significant_blocks_qvalues,
                                    self.X.index.to_numpy(),
                                    self.Y.index.to_numpy())

        # print datasets (original and discretized)
        if not self.dont_copy:
            self.X.to_csv(join(dir_name, 'X.tsv'), sep='\t', na_rep = "NA")
            self.Y.to_csv(join(dir_name, 'Y.tsv'), sep='\t', na_rep = "NA")
            self.X_ori.to_csv(join(dir_name, 'X_original.tsv'), sep='\t', na_rep = "NA")
            self.Y_ori.to_csv(join(dir_name, 'Y_original.tsv'), sep='\t', na_rep = "NA")

    '''Public functions
    '''
    def load(self, X_file, Y_file=None):
        def _read_and_drop_duplicated_indices(filepath):
            # drop duplicates and keep the first row
            df = pd.read_table(filepath, index_col=0, na_values="").dropna(how='all')
            df = df[~df.index.duplicated(keep='first')]
            return(df)

        self.logger.log_step_start('Loading and preprocessing data')
        confp = config.preprocess

        start_time = time.time()

        X, self.X_types = eval_type(_read_and_drop_duplicated_indices(X_file))
        Y, self.Y_types = eval_type(_read_and_drop_duplicated_indices(Y_file)) if Y_file \
            else (X.copy(deep=True), np.copy(self.X_types))

        # if not all types are continuous but pdist_metric is only for continuous types
        self.metric_changed = False
        if not (is_all_cont(self.X_types) and is_all_cont(self.Y_types)) and not (config.association['pdist_metric'].lower() in ['mi', 'nmi','xicor']):
            self.metric_changed = True
            self.logger.log_result('Discrete variables detected. ASSOCIATION METRIC CHANGED TO MI!', "")
            update_config('association', pdist_metric = 'mi')
            # raise ValueError('pdist_metric should be mi, nmi, or xicor if not all features are continuous...')
        # if pdist_metric is nmi but no discretization method is specified, assign to equal frequency (quantile)
        if config.association['pdist_metric'].lower() in ['nmi', 'mi'] and confp['discretize_func'] is None:
            self.logger.log_message('Discretization function is None; assigning to equal frequency (quantile) given metric in (NMI, MI)...')
            update_config('preprocess', discretize_func='quantile')
        if config.association['pdist_metric'].lower() == 'xicor' and not (is_all_cont(self.X_types) and is_all_cont(self.Y_types)) and confp['discretize_func'] is None:
            self.logger.log_message('Discretization function is None but pdist_metric = XICOR and data contains categorical variables; assigning discretization function to equal frequency (quantile)...')
            update_config('preprocess', discretize_func='quantile')
        # if all features are continuous and distance metric != nmi, discretization can be bypassed
        if is_all_cont(self.X_types) and is_all_cont(self.X_types) and confp['discretize_func'] is not None and \
            (not config.association['pdist_metric'].lower() in ['nmi', 'mi']) and confp['discretize_bypass_if_possible']:
            self.logger.log_message('All features are continuous and bypassing discretization is enabled; bypassing discretization...')
            update_config('preprocess', discretize_func=None)

        if config.association['pdist_metric'].lower() == 'nmi' and not self.force_permutations:
            self.logger.log_result('Approximating NMI p-values with chi-squared test. Use --force_permutations to disable this behavior.', "")
        if config.association['pdist_metric'].lower() == 'mi' and not self.force_permutations:
            self.logger.log_result('Approximating MI p-values with chi-squared test. Use --force_permutations to disable this behavior.', "")

        # filter tables by intersect columns
        intersect_cols = [col for col in X.columns if col in Y.columns]
        if (len(intersect_cols) < 5): self.logger.log_message("There don't seem to be many overlapping samples between the two datasets. Are you sure your datasets have features as rows and samples as columns?")
        X, Y = X[intersect_cols], Y[intersect_cols]

        # clean and preprocess data
        func_args = {
            'transform_funcs'    : confp['transform_funcs'],
            'max_freq_thresh'    : confp['max_freq_thresh'],
            'discretize_func'    : confp['discretize_func'],
            'discretize_num_bins': confp['discretize_num_bins']
        }
        self.X, self.X_ori, self.X_types = preprocess(X, self.X_types, **func_args)
        self.Y, self.Y_ori, self.Y_types = preprocess(Y, self.Y_types, **func_args)

        self.has_loaded = True
        end_time = time.time()

        self.logger.log_message('Preprocessing step completed:')
        self.logger.log_result('X shape (# features, # size)', self.X.shape)
        self.logger.log_result('Y shape (# features, # size)', self.Y.shape)
        self.logger.log_step_end('Loading and preprocessing data', end_time - start_time)

    def run(self):
        '''Run AllA:
        1) compute pairwise similarity matrix and p-values
        2) find significantly-associated cells
        '''
        if self.has_loaded == False:
            raise RuntimeError('load function has not been called!')

        self.logger.log_step_start('Performing %s' % self.name)

        # step 1: computing pairwise similarity matrix
        self._compute_pairwise_similarities()

        # step 2: find significantly-associated cells
        self._find_dense_associated_blocks()

        # generate reports
        self._generate_reports()

    def generate_hallagram(self, block_num=50, x_dataset_label='', y_dataset_label='',
                            cmap=None, cbar_label='', figsize=None, text_scale=10,
                            output_file='hallagram', mask=False, signif_dots=True, plot_type="pdf",**kwargs):
        '''Generate a hallagram showing the top [block_num] significant blocks
        '''
        if cmap is None:
            cmap = 'YlGnBu' if config.association['pdist_metric'] in ['mi','nmi', 'dcor', 'xicor'] else 'RdBu_r'
        file_name = join(config.output['dir'], output_file)
        if block_num is None:
            block_num = len(self.significant_blocks)
        else:
            block_num = min(block_num, len(self.significant_blocks))
        generate_hallagram(self.significant_blocks,
                           self.X.index.to_numpy(),
                           self.Y.index.to_numpy(),
                           [idx for idx in range(self.X.shape[0])],
                           [idx for idx in range(self.Y.shape[0])],
                           self.similarity_table,
                           self.fdr_reject_table,
                           x_dataset_label=x_dataset_label,
                           y_dataset_label=y_dataset_label,
                           figsize=figsize,
                           text_scale=text_scale,
                           output_file=file_name,
                           plot_type=plot_type,
                           cmap=cmap, cbar_label=cbar_label,
                           mask=mask, signif_dots = signif_dots, **kwargs)

########
# HAllA
########
class HAllA(AllA):
    def __init__(self, max_freq_thresh=config.preprocess['max_freq_thresh'],
                 transform_data_funcs=config.preprocess['transform_funcs'],
                 discretize_bypass_if_possible=config.preprocess['discretize_bypass_if_possible'],
                 discretize_func=config.preprocess['discretize_func'], discretize_num_bins=config.preprocess['discretize_num_bins'],
                 pdist_metric=config.association['pdist_metric'], linkage_method=config.hierarchy['linkage_method'],
                 sim2dist_set_abs=config.hierarchy['sim2dist_set_abs'], sim2dist_func=config.hierarchy['sim2dist_func'],
                 permute_func=config.permute['func'], permute_iters=config.permute['iters'], permute_speedup=config.permute['speedup'],
                 fdr_alpha=config.stats['fdr_alpha'], fdr_method=config.stats['fdr_method'],
                 fnr_thresh=config.stats['fnr_thresh'], rank_cluster=config.stats['rank_cluster'],
                 out_dir=config.output['dir'], verbose=config.output['verbose'], no_progress=False,
                 force_permutations=False, num_threads=4, dont_skip=False, large_diagnostic_subset=105, splitting_diagnostic_mode=False, gini_uncertainty_level = .02,
                 dont_copy = False, seed=0):
        # TODO: add restrictions on the input - ensure the methods specified are available
        self.name = 'HAllA'
        # retrieve AllA variables
        alla_vars = vars()
        for key in ['linkage_method', 'fnr_thresh', 'rank_cluster', 'sim2dist_set_abs', 'sim2dist_func']:
            del alla_vars[key]
        # call AllA init function
        AllA.__init__(**alla_vars)

        # update HAllA config settings
        update_config('stats', fnr_thresh=fnr_thresh, rank_cluster=rank_cluster)
        update_config('hierarchy', linkage_method=linkage_method,
                                   sim2dist_set_abs=sim2dist_set_abs,
                                   sim2dist_func=sim2dist_func)
        self.logger = HAllALogger(self.name, config=config)

    '''Private functions
    '''
    def _reset_attributes(self):
        self.X, self.Y = None, None
        self.X_types, self.Y_types = None, None
        self.X_hierarchy, self.Y_hierarchy = None, None
        self.similarity_table = None
        self.pvalue_table, self.qvalue_table = None, None
        self.fdr_reject_table = None
        self.significant_blocks = None
        self.significant_blocks_qvalues = None
        self.has_loaded = False
        self.has_run = False

    def _run_clustering(self):
        self.logger.log_step_start('Step 2: Performing hierarchical clustering', sub=True)
        start_time = time.time()
        func_args = {
            'pdist_metric'    : config.association['pdist_metric'],
            'linkage_method'  : config.hierarchy['linkage_method'],
            'sim2dist_set_abs': config.hierarchy['sim2dist_set_abs'],
            'sim2dist_func'   : config.hierarchy['sim2dist_func']
        }
        self.X_hierarchy = HierarchicalTree(self.X, **func_args)
        self.Y_hierarchy = HierarchicalTree(self.Y, **func_args)
        end_time = time.time()
        self.logger.log_step_end('Performing hierarchical clustering', end_time - start_time, sub=True)

    def _find_dense_associated_blocks(self):
        def sort_by_best_qvalue(x):
            qvalue_table = self.qvalue_table[x[0],:][:,x[1]]
            return(qvalue_table.min())
        def sort_by_avg_qvalue(x):
            qvalue_table = self.qvalue_table[x[0],:][:,x[1]]
            return(qvalue_table.mean())

        self.logger.log_step_start('Step 3: Finding densely associated blocks', sub=True)
        start_time = time.time()
        self.significant_blocks = compare_and_find_dense_block(self.X_hierarchy.tree, self.Y_hierarchy.tree,
                                     self.fdr_reject_table, fnr_thresh=config.stats['fnr_thresh'], 
                                     splitting_diagnostic_mode = self.splitting_diagnostic_mode,
                                     gini_uncertainty_level=self.gini_uncertainty_level)
        # sort significant blocks by the rank_cluster method
        sort_func = sort_by_best_qvalue if config.stats['rank_cluster'] == 'best' else sort_by_avg_qvalue
        self.significant_blocks.sort(key=sort_func)
        self.significant_blocks_qvalues = [sort_func(x) for x in self.significant_blocks]
        end_time = time.time()
        self.logger.log_result('Number of significant clusters', len(self.significant_blocks))
        self.logger.log_step_end('Finding densely associated blocks', end_time - start_time, sub=True)

    def _generate_reports(self):
        '''Generate reports and store in config.output['dir'] directory
        '''
        AllA._generate_reports(self)

        # store the linkages in .npy files
        dir_name = config.output['dir']
        np.save(join(dir_name, 'X_linkage.npy'), self.X_hierarchy.linkage, allow_pickle=False)
        np.save(join(dir_name, 'Y_linkage.npy'), self.Y_hierarchy.linkage, allow_pickle=False)

    '''Public functions
    '''
    def run(self):
        '''Run all 3 steps:
        1) compute pairwise similarity matrix
        2) cluster hierarchically
        3) find densely-associated blocks iteratively
        '''
        if self.has_loaded == False:
            raise RuntimeError('load function has not been called!')

        self.logger.log_step_start('Performing %s' % self.name)

        # step 1: computing pairwise similarity matrix
        self._compute_pairwise_similarities()

        # step 2: hierarchical clustering
        self._run_clustering()
        # step 3: iteratively finding densely-associated blocks
        self._find_dense_associated_blocks()

        # generate reports
        self._generate_reports()

    def generate_hallagram(self, block_num=50, x_dataset_label='', y_dataset_label='',
                            cmap=None, cbar_label='', figsize=None, text_scale=10,
                            output_file='hallagram', mask=False, signif_dots=True, plot_type="pdf", **kwargs):
        '''Generate a hallagram showing the top [block_num] significant blocks
        '''
        if cmap is None:
            cmap = 'YlGnBu' if config.association['pdist_metric'] in ['mi', 'nmi', 'dcor', 'xicor'] else 'RdBu_r'
        file_name = join(config.output['dir'], output_file)
        if block_num is None:
            block_num = len(self.significant_blocks)
        else:
            block_num = min(block_num, len(self.significant_blocks))
        generate_hallagram(self.significant_blocks,
                           self.X.index.to_numpy(),
                           self.Y.index.to_numpy(),
                           self.X_hierarchy.tree.pre_order(),
                           self.Y_hierarchy.tree.pre_order(),
                           self.similarity_table,
                           fdr_reject_table=self.fdr_reject_table,
                           x_dataset_label=x_dataset_label,
                           y_dataset_label=y_dataset_label,
                           figsize=figsize,
                           text_scale=text_scale,
                           plot_type=plot_type,
                           output_file=file_name,
                           cmap=cmap, cbar_label=cbar_label,
                           mask=mask, signif_dots=signif_dots, block_num=block_num, **kwargs)

    def generate_clustermap(self, x_dataset_label='', y_dataset_label='',
                            cmap=None, cbar_label='', figsize=None, text_scale=10,
                            output_file='clustermap.pdf', mask=False, signif_dots=True, **kwargs):
        '''Generate a clustermap (hallagram + dendrogram)
        '''
        # if the dimension is too large, generate a hallagram instead
        if max(self.similarity_table.shape) > 500:
            print('The dimension is too large - please generate a hallagram instead.', file = sys.stderr)
            return
        if cmap is None:
            cmap = 'YlGnBu' if config.association['pdist_metric'] in ['mi','nmi', 'dcor', 'xicor'] else 'RdBu_r'

        file_name = join(config.output['dir'], output_file)
        generate_clustermap(self.significant_blocks,
                            self.X.index.to_numpy(),
                            self.Y.index.to_numpy(),
                            self.X_hierarchy.linkage,
                            self.Y_hierarchy.linkage,
                            self.similarity_table,
                            fdr_reject_table=self.fdr_reject_table,
                            x_dataset_label=x_dataset_label,
                            y_dataset_label=y_dataset_label,
                            figsize=figsize,
                            text_scale=text_scale,
                            cmap=cmap, cbar_label=cbar_label,
                            output_file=file_name,
                            mask=mask,
                            signif_dots=signif_dots,
                            **kwargs)

    def generate_diagnostic_plot(self, block_num=50, plot_dir='diagnostic', axis_stretch=1e-5, plot_size=4):
        '''Generate a lattice plot for each significant association;
        save all plots in the plot_dir folder under config.output['dir']
        '''
        # create the diagnostic directory under config.output['dir']
        reset_dir(join(config.output['dir'], plot_dir))
        if block_num is None:
            block_num = len(self.significant_blocks)
        else:
            block_num = min(block_num, len(self.significant_blocks))
        for i, block in enumerate(self.significant_blocks[:block_num]):
            title = 'Association %d' % (i+1)
            out_file = join(config.output['dir'], plot_dir, 'association_%d.pdf' % (i+1))
            warn_file = join(config.output['dir'], plot_dir, 'warnings.txt')
            x_data = self.X.to_numpy()[block[0],:]
            y_data = self.Y.to_numpy()[block[1],:]
            x_ori_data = self.X_ori.to_numpy()[block[0],:]
            y_ori_data = self.Y_ori.to_numpy()[block[1],:]
            x_features = self.X.index.to_numpy()[block[0]]
            y_features = self.Y.index.to_numpy()[block[1]]
            x_types = np.array(self.X_types)[block[0]]
            y_types = np.array(self.Y_types)[block[1]]
            if (x_data.shape[0] + y_data.shape[0]) > 15 and (x_data.shape[0] + y_data.shape[0]) <= 45:
                warn_string = "Over 15 features included in association %d. Only a subset of features will be shown in the diagnostic plot. Increase --large_diagnostic_subset beyond 105 to show more." % (i+1)
                if exists(warn_file):
                    append_write = 'a'
                else:
                    append_write = 'w'
                warn_file_write = open(warn_file, append_write)
                warn_file_write.write(warn_string + '\n')
                warn_file_write.close()
                print(warn_string, file = sys.stderr)
                generate_lattice_plot(x_data, y_data, x_ori_data, y_ori_data,
                                    x_features, y_features, x_types, y_types, title,
                                    out_file, axis_stretch=axis_stretch, plot_size=plot_size, n_pairs_to_show = self.large_diagnostic_subset)
                continue
            if (x_data.shape[0] + y_data.shape[0]) > 45 and not self.dont_skip:
                warn_string = "Skipping association %d because there are too many included features. Add --dont_skip_large_blocks to disable this behavior." % (i+1)
                if exists(warn_file):
                    append_write = 'a'
                else:
                    append_write = 'w'
                warn_file_write = open(warn_file, append_write)
                warn_file_write.write(warn_string + '\n')
                warn_file_write.close()
                print(warn_string, file = sys.stderr)
                continue
            generate_lattice_plot(x_data, y_data, x_ori_data, y_ori_data,
                                    x_features, y_features, x_types, y_types, title,
                                    out_file, axis_stretch=axis_stretch, plot_size=plot_size, n_pairs_to_show = (x_data.shape[0] + y_data.shape[0])**2)



from .similarity import does_return_pval, get_similarity_function, remove_missing_values

import numpy as np
import sys
import scipy.spatial.distance as spd
from statsmodels.stats.multitest import multipletests
from scipy.stats import genpareto
import itertools
from multiprocessing import Pool, cpu_count
from tqdm import tqdm
from time import time

# retrieve package named 'eva' from R for GPD-related calculations
from rpy2.robjects.packages import importr
from rpy2.robjects.vectors import FloatVector
eva = importr('eva')

'''P-value computations by permutation test
'''
def compute_pvalue_ecdf(permuted_scores, gt_score, n):
    '''Compute the right-tailed test p-value using empirical cumulative distribution function given
    - permuted_scores: the scores computed by deriving association between x and shuffled y
    - gt_score       : the test statistic
    - n              : the number of iterations
    '''
    permuted_scores = np.array(permuted_scores)
    gt_score = np.abs(gt_score)
    pval = ((permuted_scores > gt_score).sum() + 1) / n
    return(min(1.0, pval))

def compute_pvalue_gpd(permuted_scores, gt_score, n):
    '''Approximate the p-value using generalized pareto distribution given
    Code converted to python from R: https://github.com/goncalves-lab/waddR/blob/master/R/PValues.R
    - permuted_scores: the scores computed by deriving association between x and shuffled y
    - gt_score       : the test statistic
    - n              : the number of iterations
    '''

    def get_pvalue(sorted_scores, stat, n):
        # approximate the gpd tail
        n_exceed = 250
        is_gpd_fitted = False
        while n_exceed >= 10:
            exceedances = sorted_scores[:n_exceed]
            # check if the n_exceed largest permutation values follow GPD
            #   with Anderson-Darling goodness-of-fit test
            try:
                ad = eva.gpdAd(FloatVector(exceedances))
                ad_pval = ad.rx2('p.value')[0]
            except:
                n_exceed -= 10
                continue
            # H0 = exceedances come from a GPD
            if ad_pval > 0.05:
                is_gpd_fitted = True
                break
            n_exceed -= 10
        if not is_gpd_fitted:
            #print('GPD good fit is never reached - use ECDF instead...')
            return(None)
        # compute the exceedance threshold t
        t = float((sorted_scores[n_exceed] + sorted_scores[n_exceed-1])/2)
        # estimate shape and scale params with maximum likelihood
        gpd_fit = eva.gpdFit(FloatVector(sorted_scores), threshold=t, method='mle')
        scale, shape = gpd_fit.rx2('par.ests')[0], gpd_fit.rx2('par.ests')[1]

        # compute GPD p-value
        f_gpd = genpareto.cdf(x=gt_score-t, c=shape, scale=scale)
        return(n_exceed / n * (1 - f_gpd))

    gt_score = abs(gt_score)
    # approximate the tail
    sorted_scores = sorted(permuted_scores, reverse=True)
    return(get_pvalue(sorted_scores, gt_score, n))

def compute_permutation_test_pvalue(x, y, pdist_metric='nmi', permute_func='gpd',
                                    iters=10000, speedup=True, alpha=0.05, seed=None):
    '''Compute two-sided p-value using permutation test of the pairwise similarity between
        an X feature and a Y feature by:
        - set all scores and test statistic to be absolute values
        - do right-tailed test
    '''
    def _compute_score(feat1, feat2):
        score = spd.cdist(feat1, feat2, metric=get_similarity_function(pdist_metric))
        return(np.abs(score[0,0])) # original shape is (1,1)

    if seed:
        np.random.seed(seed)
    if len(x.shape) == 1: x = x.reshape((1, len(x)))
    if len(y.shape) == 1: y = y.reshape((1, len(y)))
    permute_func = permute_func.lower()
    permuted_dist_scores = []
    # compute the ground truth scores for comparison later
    rmx, rmy = remove_missing_values(x,y)
    if (np.unique(rmx).shape[0] == 1 or np.unique(rmy).shape[0] == 1):
        return(1)
    gt_score = _compute_score(x, y)
    permuted_y = np.copy(y[0])
    best_pvalue = 1.0
    permutation_num = iters
    for iter in range(iters):
        np.random.shuffle(permuted_y)
        # compute permuted score and append to the list
        permuted_dist_scores.append(_compute_score(x, permuted_y.reshape(y.shape)))
        if (iter+1) % 100 == 0 and speedup:
            curr_pvalue = compute_pvalue_ecdf(permuted_dist_scores, gt_score, iter+1)
            if curr_pvalue <= best_pvalue:
                best_pvalue = curr_pvalue
            elif curr_pvalue > alpha and (iter+1) >= 300:
                # only break if curr_pvalue > best_pvalue, curr_pvalue > alpha, iters >= 300 (arbitrary)
                permutation_num = iter + 1
                break
    if permute_func == 'ecdf': # empirical cumulative dist. function
        return(compute_pvalue_ecdf(permuted_dist_scores, gt_score, permutation_num))
    # gpd algorithm - Knijnenburg2009, Ge2012
    # compute M - # null samples exceeding the test statistic
    # recall that gt_score is positive
    M = len([1 for score in permuted_dist_scores if score > gt_score])
    # if M >= 10, use ecdf
    if M >= 10:
        return(compute_pvalue_ecdf(permuted_dist_scores, gt_score, permutation_num))

    if speedup:
        # attempt to use gpd
        pval = compute_pvalue_gpd(permuted_dist_scores, gt_score, permutation_num)
        if (pval is None):
            pval = compute_pvalue_ecdf(permuted_dist_scores, gt_score, permutation_num)
    else:
        pval = compute_pvalue_ecdf(permuted_dist_scores, gt_score, permutation_num)

    return(pval)

def pval_only(x,y, pval_fun, return_pval=True):
    return(pval_fun(x,y,return_pval)[1])

def get_pvalue_table(X, Y, pdist_metric='nmi', permute_func='gpd', permute_iters=1000,
                     permute_speedup=True, alpha=0.05, seed=None, no_progress=False, force_permutations=False, num_threads=4):
    '''Obtain pairwise p-value tables given features in X and Y
    '''
    # initiate table
    n, m = X.shape[0], Y.shape[0]
    pvalue_table = np.zeros((n, m))
    if does_return_pval(pdist_metric) and (not force_permutations) and num_threads == 1:
        for i in tqdm(range(n), disable=no_progress):
            for j in range(m):
                pvalue_table[i,j] = get_similarity_function(pdist_metric)(X[i,:], Y[j,:], return_pval=True)[1]
    elif does_return_pval(pdist_metric) and (not force_permutations):
        with Pool(processes=num_threads) as pool:
            pvalue_table = pool.starmap(pval_only, [(X[i,:], Y[j,:], get_similarity_function(pdist_metric), True)\
                                                                                  for i in range(n) for j in range(m)])
            pvalue_table = np.array(pvalue_table).reshape((n, m))
    else:
        with Pool(processes=num_threads) as pool:
            pvalue_table = pool.starmap(compute_permutation_test_pvalue, [(X[i,:], Y[j,:], pdist_metric,
                                                                            permute_func, permute_iters,
                                                                            permute_speedup, alpha, seed)\
                                                                            for i in range(n) for j in range(m)])
            pvalue_table = np.array(pvalue_table).reshape((n, m))
    return(pvalue_table)

def test_pvalue_run_time(X, Y, pdist_metric='nmi', permute_func='gpd', permute_iters=1000,
                     permute_speedup=True, alpha=0.05, force_perms=False, num_threads=4, seed=None):
    '''
    Run a p-value computation test and return the time it took and a message extrapolating to the full dataset.
    '''
    test_start = time()

    if does_return_pval(pdist_metric) and not force_perms:
        get_similarity_function(pdist_metric)(X[0,:], Y[0,:], return_pval=True)[1]
        n_threads = 1
    else:
        n_threads = num_threads
        compute_permutation_test_pvalue(X[0,:], Y[0,:],
                                    pdist_metric=pdist_metric,
                                    permute_func=permute_func,
                                    iters=permute_iters,
                                    speedup=permute_speedup, alpha=alpha, seed=seed)

    test_end = time()
    test_length = test_end - test_start
    extrapolated_time = 4 * test_length * X.shape[0] * Y.shape[0] / n_threads # 4 = safety factor
    timing_string = "The first p-value computation took about " + str(round(test_length, 3)) + " seconds. Extrapolating from this, computing the entire p-value table should take roughly " + str(round(extrapolated_time,3)) + " seconds..."
    return(extrapolated_time, timing_string)

def pvalues2qvalues(pvalues, method='fdr_bh', alpha=0.05):
    '''Perform p-value correction for multiple tests (Benjamini/Hochberg)
    Args:
    - pvalues: a 1D-array of pvalues
    - alpha  : family-wise error rate
    Return a tuple (adjusted p-value array, boolean array [True = reject]
    '''
    return(multipletests(pvalues, alpha=alpha, method=method)[:2])

def compute_result_power(significant_blocks, true_assoc):
    '''Compute power (recall: TP / condition positive) given args:
    - significant blocks: a list of significant blocks in the original indices, e.g.,
                          [[[2], [0]], [[0,1], [1]]] --> two blocks
    - true_assoc        : A matrix with row ~ X features and col ~ Y features containing
                          1 if association exists or 0 if not
    '''
    positive_cond = np.sum(true_assoc).astype(int)
    positive_true = 0
    for block in significant_blocks:
        for i,j in itertools.product(block[0], block[1]):
            if true_assoc[i][j] == 1: positive_true += 1
    return(positive_true * 1.0 / positive_cond)

def compute_result_fdr(significant_blocks, true_assoc):
    '''Compute fdr (FP / predicted condition positive) given args:
    - significant blocks: a list of significant blocks in the original indices, e.g.,
                          [[[2], [0]], [[0,1], [1]]] --> two blocks
    - true_assoc        : A matrix with row ~ X features and col ~ Y features containing
                          1 if association exists or 0 if not
    '''
    false_positive = 0
    predicted_positive = 0
    for block in significant_blocks:
        for i,j in itertools.product(block[0], block[1]):
            predicted_positive += 1
            if true_assoc[i][j] != 1: false_positive += 1
    if predicted_positive == 0: return(np.nan)
    return(false_positive * 1.0 / predicted_positive)



import unittest
import sys
from os.path import dirname, abspath
import numpy as np
from scipy.stats import spearmanr, pearsonr
from halla.utils import stats

from utils import compare_numpy_array

class TestStatsUtils(unittest.TestCase):
    '''Tests the p-value permutation test function; extreme cases
    '''
    def test_compute_permutation_test_pvalue_significant(self):
        np.random.seed(1)
        eps = 0.001
        # source: https://machinelearningmastery.com/how-to-calculate-nonparametric-rank-correlation-in-python/
        x = np.random.normal(size=1000) * 20
        y = x + np.random.normal(size=1000) * 10
        # get expected p-value
        _, expected_pvalue = spearmanr(x, y)
        test_pvalue = stats.compute_permutation_test_pvalue(x, y, pdist_metric='spearman',
                        permute_func='ecdf', speedup=False, iters=1000, seed=123)
        self.assertLessEqual(abs(test_pvalue - expected_pvalue), eps)
    
    def test_compute_permutation_test_pvalue_significant_speedup(self):
        np.random.seed(1)
        eps = 0.001
        # source: https://machinelearningmastery.com/how-to-calculate-nonparametric-rank-correlation-in-python/
        x = np.random.normal(size=1000) * 20
        y = x + np.random.normal(size=1000) * 10
        # get expected p-value
        _, expected_pvalue = spearmanr(x, y)
        test_pvalue = stats.compute_permutation_test_pvalue(x, y, pdist_metric='spearman',
                        permute_func='ecdf', speedup=True, alpha=0.05, iters=1000, seed=123)
        self.assertLessEqual(abs(test_pvalue - expected_pvalue), eps)
    
    def test_compute_permutation_test_pvalue_insignificant(self):
        np.random.seed(2)
        eps = 0.02
        x = np.random.normal(size=1000)
        y = np.random.normal(size=1000)
        _, expected_pvalue  = pearsonr(x,y)
        test_pvalue = stats.compute_permutation_test_pvalue(x, y, pdist_metric='pearson',
                        permute_func='ecdf', speedup=False, iters=1000, seed=123)
        self.assertLessEqual(abs(test_pvalue - expected_pvalue), eps)

    def test_compute_permutation_test_pvalue_insignificant_speedup(self):
        np.random.seed(2)
        x = np.random.normal(size=1000)
        y = np.random.normal(size=1000)
        test_pvalue = stats.compute_permutation_test_pvalue(x, y, pdist_metric='pearson',
                        permute_func='ecdf', speedup=True, alpha=0.05, iters=1000, seed=123)
        # we don't care about the exact value
        self.assertTrue(test_pvalue > 0.05)
    
'''Other cases; programmatically create tests as attributes to TestStatsUtils
'''
def test_generator_ecdf(mu_pair, var_pair, corr, sample_size):
    np.random.seed(123)
    eps = 0.03
    cov_matrix = np.array([
                    [var_pair[0], corr],
                    [corr, var_pair[1]]
                ])
    x, y = np.random.multivariate_normal(mu_pair, cov_matrix, sample_size).T
    def test(self):
        _, expected_pvalue = spearmanr(x, y)
        test_pvalue_nospeedup = stats.compute_permutation_test_pvalue(x, y, pdist_metric='spearman',
                                permute_func='ecdf', speedup=False, iters=1000, seed=123)
        self.assertLessEqual(abs(test_pvalue_nospeedup - expected_pvalue), eps)
    return test

def test_generator_gpd(mu_pair, var_pair, corr, sample_size):
    np.random.seed(123)
    eps = 0.005
    cov_matrix = np.array([
                    [var_pair[0], corr],
                    [corr, var_pair[1]]
                ])
    x, y = np.random.multivariate_normal(mu_pair, cov_matrix, sample_size).T
    def test(self):
        _, expected_pvalue = spearmanr(x, y)
        test_pvalue = stats.compute_permutation_test_pvalue(x, y, pdist_metric='spearman',
                        permute_func='gpd', speedup=True, alpha=0.05, iters=1000, seed=123)
        if expected_pvalue <= 0.05: # significant
            self.assertLessEqual(abs(test_pvalue - expected_pvalue), eps)
        else:
            self.assertTrue(test_pvalue > 0.05)
    return test

corrs = [0.8, 0.03, -0.15, -0.4]
sample_sizes = [10, 50, 100]
mu_pairs = [[0, 0]]
variance_pairs = [[10.7, 20.3], [1, 50], [1,1]]

counter = 0
for corr in corrs:
    for sample_size in sample_sizes:
        for mu_pair in mu_pairs:
            for var_pair in variance_pairs:
                # ecdf
                counter += 1
                test_name = 'test_compute_permutation_test_pvalue_ecdf_case%s' % counter
                test = test_generator_ecdf(mu_pair, var_pair, corr, sample_size)
                setattr(TestStatsUtils, test_name, test)

                # gpd
                counter += 1
                test_name = 'test_compute_permutation_test_pvalue_gpd_case%s' % counter
                test = test_generator_gpd(mu_pair, var_pair, corr, sample_size)
                setattr(TestStatsUtils, test_name, test)


import matplotlib.pyplot as plt
import matplotlib.patheffects as path_effects
import seaborn as sns
import numpy as np
import scipy.cluster.hierarchy as sch
import pandas as pd
from os.path import join
import itertools
import sys
from mpl_toolkits.axes_grid1 import make_axes_locatable
from copy import deepcopy

def get_indices_map_dict(new_indices):
    return({ idx: i for i, idx in enumerate(new_indices) })

def get_included_features(significant_blocks, num_x_features, num_y_features, trim=True, forced_x_idx=None, forced_y_idx=None):
    '''if trim is True, returns only the included features in X and Y. Optionally force certain features to be included.
    '''
    if trim:
        included_x_features, included_y_features = [], []
        for block in significant_blocks:
            included_x_features = included_x_features + block[0]
            included_y_features = included_y_features + block[1]
        for forced_x in forced_x_idx:
            included_x_features = included_x_features + forced_x_idx
        for forced_y in forced_y_idx:
            included_y_features = included_y_features + forced_y_idx
        included_x_features = sorted(list(set(included_x_features)))
        included_y_features = sorted(list(set(included_y_features)))
    else:
        included_x_features = [idx for idx in range(num_x_features)]
        included_y_features = [idx for idx in range(num_y_features)]
    return(included_x_features, included_y_features)

def remove_unshown_features(significant_blocks, shown_x, shown_y):
    '''
    Given a set of significant blocks and features that will be shown, go through each block and remove features that won't be shown

    '''
    shown_x_set = set(shown_x)
    shown_y_set = set(shown_y)
    signif_blocks = deepcopy(significant_blocks)
    has_deleted = np.zeros([len(signif_blocks),2],
                           dtype = bool)
    for block in range(len(signif_blocks)):
        if (not set(signif_blocks[block][0]).issubset(shown_x_set)):
            has_deleted[block,0] = True
        if (not set(signif_blocks[block][1]).issubset(shown_y_set)):
            has_deleted[block,1] = True
        signif_blocks[block][0] = list(set(signif_blocks[block][0]).intersection(shown_x_set))
        signif_blocks[block][1] = list(set(signif_blocks[block][1]).intersection(shown_y_set))

    to_delete = []
    for i in range(len(signif_blocks)):
        if (not signif_blocks[i][0] or not signif_blocks[i][1]) or (len(signif_blocks[i][0]) == 1 and len(signif_blocks[i][1]) == 1):
            to_delete.append(i)

    for i in sorted(to_delete, reverse=True):
        del signif_blocks[i]
        np.delete(has_deleted, i, axis = 0)

    return(signif_blocks, has_deleted)

def generate_hallagram(significant_blocks, x_features, y_features, clust_x_idx, clust_y_idx, sim_table, fdr_reject_table,
                        x_dataset_label='', y_dataset_label='', mask=False, trim=True,
                        signif_dots=True, block_num=50, show_lower=True, force_x_ft=None, force_y_ft=None, dpi=100, suppress_numbers=False,
                        figsize=None, cmap='RdBu_r', cbar_label='', text_scale=10, block_border_width=1.65, output_file='out', plot_type="pdf", **kwargs):
    '''Plot hallagram given args:
    - significant blocks: a list of *ranked* significant blocks in the original indices, e.g.,
                          [[[2], [0]], [[0,1], [1]]] --> two blocks
    - {x,y}_features     : feature names of {x,y}
    - clust_{x,y}_idx    : the indices of {x,y} in clustered form
    - sim_table          : similarity table with size [len(x_features), len(y_features)]
    - fdr_reject_table   : boolean association table with size [len(x_features), len(y_features)]
    - {x,y}_dataset_label: axis label
    - mask               : if True, mask all cells not included in significant blocks
    - trim               : if True, trim all features that are not significant
    - signif_dots        : if True, show dots on significant pairwise associations
    - suppress_numbers   : if True, don't number the blocks
    - block_num          : number indicating the top N blocks to include
    - show_lower         : if True, put a grey box around blocks ranked below the block_num threshold that show up on the hallagram anyway
    - figsize            : figure size
    - cmap               : color map
    - text_scale         : how much the rank text size should be scaled
    - block_border_width : the border width for all blocks
    - output_file        : file path to store the hallagram
    - kwargs             : other keyword arguments to be passed to seaborn's heatmap()
    '''
    #---data preparation---#

    if trim:
        top_blocks = significant_blocks[:block_num] # these are the first N blocks that MUST be shown & highlighted
    else:
        top_blocks = significant_blocks

    if len(top_blocks) == 0:
        print('The length of significant blocks is 0, no hallagram can be generated...', file = sys.stderr)
        return

    if force_x_ft is None:
        force_x_ft = []
    if force_y_ft is None:
        force_y_ft = []

    if force_x_ft:
        forced_x_idx = [x_features.tolist().index(i) for i in force_x_ft]
    else:
        forced_x_idx = []
    if force_y_ft:
        forced_y_idx = [y_features.tolist().index(i) for i in force_y_ft]
    else:
        forced_y_idx = []

    included_x_feat, included_y_feat = get_included_features(top_blocks,
                                                             len(x_features),
                                                             len(y_features),
                                                             trim,
                                                             forced_x_idx,
                                                             forced_y_idx)
    if show_lower:
        lower_blocks, has_deleted = remove_unshown_features(significant_blocks[block_num:],
                                                            included_x_feat,
                                                            included_y_feat)

    # filter the indices with the included features
    clust_x_idx = np.asarray([i for i in clust_x_idx if i in included_x_feat])
    clust_y_idx = np.asarray([i for i in clust_y_idx if i in included_y_feat])

    # if mask, replace all insignificant cells to NAs
    clust_sim_table = np.copy(sim_table)
    clust_fdr_reject_table = np.copy(fdr_reject_table)
    if mask:
        dummy_table = np.full(clust_sim_table.shape, np.nan)
        dummy_fdr = np.full(clust_sim_table.shape, np.nan)
        for block in top_blocks:
            for x, y in itertools.product(block[0], block[1]):
                dummy_table[x,y] = clust_sim_table[x,y]
                dummy_fdr[x,y] = clust_fdr_reject_table[x,y]
        clust_sim_table = dummy_table
        clust_fdr_reject_table = dummy_fdr

    # shuffle similarity table and features accordingly
    clust_sim_table = clust_sim_table[clust_x_idx,:][:,clust_y_idx]
    clust_fdr_reject_table = clust_fdr_reject_table[clust_x_idx,:][:,clust_y_idx]
    clust_x_features = np.asarray(x_features)[clust_x_idx]
    clust_y_features = np.asarray(y_features)[clust_y_idx]

    # create a dict to ease indices conversion
    x_ori2clust_idx = get_indices_map_dict(clust_x_idx)
    y_ori2clust_idx = get_indices_map_dict(clust_y_idx)

    #---begin plotting---#
    # set up colormap anchor
    vmax, vmin = np.max(clust_sim_table), np.min(clust_sim_table)
    if vmin < 0 and vmax > 0:
        vmax = max(abs(vmin), vmax)
        vmin = -vmax

    sns.set_style('whitegrid')
    if figsize is None:
        # set each cell as ~0.3, then add margins
        figsize = (max(5, 0.3*clust_sim_table.shape[1]+4), max(5, 0.3*clust_sim_table.shape[0]+4))

    _, ax = plt.subplots(figsize=figsize)
    # add colorbar axes
    divider = make_axes_locatable(ax)
    cbar_ax = divider.append_axes('left', size=str(100/(figsize[0]*2.5))+'%', pad=1/(figsize[0]*6))
    cbar_kws = { 'label': cbar_label, 'ticklocation': 'left' }
    sns.heatmap(clust_sim_table, xticklabels = clust_y_features,
                        cmap=cmap, vmin=vmin, vmax=vmax, square=True,
                        zorder=3, cbar_ax=cbar_ax, ax=ax, cbar_kws=cbar_kws, **kwargs)
    ax.yaxis.tick_right()
    ax.yaxis.set_label_position('right')
    ax.tick_params(axis = 'y', length=0)
    ax.set_yticklabels(clust_x_features, rotation=0, ha='left')
    if mask:
        # minor ticks
        ax.set_xticks(np.arange(0, clust_sim_table.shape[1], 1), minor=True)
        ax.set_yticks(np.arange(0, clust_sim_table.shape[0], 1), minor=True)
        ax.grid(which='minor', color='xkcd:light grey', zorder=0)
    ax.set_xlabel(y_dataset_label, fontweight='bold')
    ax.set_ylabel(x_dataset_label, fontweight='bold')

    if signif_dots:
        for i in range(len(clust_y_features)):
            for j in range(len(clust_x_features)):
                if clust_fdr_reject_table[j,i]:
                    ax.scatter(x = i+.5, y =j + .5, c = 'black', marker = "o", zorder = 3, s=25)
                    ax.scatter(x = i+.5, y =j + .5, c = 'white', marker = "o", zorder = 3, s=10)

    if show_lower:
        block_i = 0
        for rank, block in enumerate(lower_blocks):
            x_block, y_block = block[0], block[1]
            clust_x_block = [x_ori2clust_idx[idx] for idx in x_block]
            clust_y_block = [y_ori2clust_idx[idx] for idx in y_block]
            if has_deleted[block_i,0]:
                ax.vlines([min(clust_y_block)], min(clust_x_block), max(clust_x_block)+1, color='0.2', linewidths=block_border_width, zorder=4, alpha = .5, capstyle="projecting")
                ax.vlines([max(clust_y_block)+1], min(clust_x_block), max(clust_x_block)+1, color='0.2', linewidths=block_border_width, zorder=4, linestyles='dotted', alpha = .5, capstyle="projecting")
            else:
                ax.vlines([min(clust_y_block), max(clust_y_block)+1], min(clust_x_block), max(clust_x_block)+1, color='0.2', linewidths=block_border_width, zorder=4, alpha = .5, capstyle="projecting")

            if has_deleted[block_i,1]:
                ax.hlines([min(clust_x_block)], min(clust_y_block), max(clust_y_block)+1, color='0.2', linewidths=block_border_width, zorder=4, alpha = .5, capstyle="projecting")
                ax.hlines([max(clust_x_block)+1], min(clust_y_block), max(clust_y_block)+1, color='0.2', linewidths=block_border_width, zorder=4, linestyles='dotted', alpha = .5, capstyle="projecting")
            else:
                ax.hlines([min(clust_x_block), max(clust_x_block)+1], min(clust_y_block), max(clust_y_block)+1, color='0.2', linewidths=block_border_width, zorder=4, alpha = .5, capstyle="projecting")
            block_i += 1

    for rank, block in enumerate(top_blocks):
        x_block, y_block = block[0], block[1]
        clust_x_block = [x_ori2clust_idx[idx] for idx in x_block]
        clust_y_block = [y_ori2clust_idx[idx] for idx in y_block]
        ax.vlines([min(clust_y_block), max(clust_y_block)+1], min(clust_x_block), max(clust_x_block)+1, color='black', linewidths=block_border_width, zorder=4, capstyle="projecting")
        ax.hlines([min(clust_x_block), max(clust_x_block)+1], min(clust_y_block), max(clust_y_block)+1, color='black', linewidths=block_border_width, zorder=4, capstyle="projecting")
        # add rank text
        if not suppress_numbers:
            text_content = str(rank + 1)
            text_size = (min(max(clust_y_block) - min(clust_y_block),
                             max(clust_x_block) - min(clust_x_block)) + 1)*text_scale
            text = ax.text(
                np.mean(clust_y_block) + 0.5, np.mean(clust_x_block) + 0.5,
                text_content, size=text_size, color='white', ha='center', va='center', weight='bold')
            text.set_path_effects([
                path_effects.Stroke(linewidth=3, foreground='black'),
                path_effects.Normal(),
            ])

    if not ("." in plot_type):
        plot_type = "." + plot_type
        
    plt.subplots_adjust(wspace=1/(figsize[0]*6), hspace=0)
    plt.savefig(output_file + plot_type, bbox_inches='tight', dpi=dpi)

def generate_clustermap(significant_blocks, x_features, y_features, x_linkage, y_linkage, sim_table, fdr_reject_table,
                        x_dataset_label='', y_dataset_label='', signif_dots=True, figsize=None, cmap='RdBu_r', text_scale=10,
                        dendrogram_ratio=None, cbar_label='',
                        block_border_width=1.5, mask=False, output_file='out.pdf', **kwargs):
    '''Plot a clustermap given args:
    - significant blocks: a list of *ranked* significant blocks in the original indices, e.g.,
                          [[[2], [0]], [[0,1], [1]]] --> two blocks
    - {x,y}_features     : feature names of {x,y}
    - {x,y}_linkage      : precomputed linkage matrix for {x,y}
    - sim_table          : similarity table with size [len(x_features), len(y_features)]
    - fdr_reject_table   : boolean association table with size [len(x_features), len(y_features)]
    - {x,y}_dataset_label: axis label
    - signif_dots        : if True, show dots on significant pairwise associations
    - figsize            : figure size
    - cmap               : color map
    - text_scale         : how much the rank text size should be scaled
    - block_border_width : the border width for all blocks
    - mask               : if True, mask all cells not included in significant blocks
    - output_file        : file path to store the hallagram
    - kwargs             : other keyword arguments to be passed to seaborn's clustermap()
    '''
    vmax, vmin = np.max(sim_table), np.min(sim_table)
    if vmin < 0 and vmax > 0:
        vmax = max(abs(vmin), vmax)
        vmin = -vmax
    mask_ar = None
    if mask:
        sns.set_style('white')
        mask_ar = np.full(sim_table.shape, True, dtype=bool)
        for block in significant_blocks:
            for i, j in itertools.product(block[0], block[1]):
                mask_ar[i][j] = False
    if figsize is None:
        figsize = (max(5, 0.3*sim_table.shape[1]+2.5), max(5, 0.3*sim_table.shape[0]+4))
    if dendrogram_ratio is None:
        dendrogram_ratio = (0.8/figsize[0], 0.8/figsize[1])
    cbar_pos = (0, 0.8, 0.1/figsize[0], 0.18)
    cbar_kws = { 'label': cbar_label, 'ticklocation': 'left' }
    clustermap = sns.clustermap(sim_table, row_linkage=x_linkage, col_linkage=y_linkage, cmap=cmap, mask=mask_ar, zorder=2,
                        xticklabels=y_features, yticklabels=x_features, vmin=vmin, vmax=vmax, cbar_pos=cbar_pos, cbar_kws=cbar_kws,
                        figsize=figsize, dendrogram_ratio=dendrogram_ratio, **kwargs)
    ax = clustermap.ax_heatmap
    ax.set_xlabel(y_dataset_label, fontweight='bold')
    ax.set_ylabel(x_dataset_label, fontweight='bold')
    if mask:
        # minor ticks
        ax.set_xticks(np.arange(0, sim_table.shape[1], 1), minor=True)
        ax.set_yticks(np.arange(0, sim_table.shape[0], 1), minor=True)
        ax.grid(which='minor', color='xkcd:light grey', zorder=0)
    x_ori2clust_idx = get_indices_map_dict(np.asarray(sch.to_tree(x_linkage).pre_order()))
    y_ori2clust_idx = get_indices_map_dict(np.asarray(sch.to_tree(y_linkage).pre_order()))

    dot_order_x = np.asarray(clustermap.dendrogram_row.reordered_ind)
    dot_order_y = clustermap.dendrogram_col.reordered_ind

    if signif_dots:
        for i in range(len(x_features)):
            for j in range(len(y_features)):
                if fdr_reject_table[dot_order_x[i],dot_order_y[j]]:
                    ax.scatter(y = i + .5, x = j + .5, c = 'black', marker = "o", zorder = 3, s = 25)
                    ax.scatter(y = i + .5, x = j + .5, c = 'white', marker = "o", zorder = 3, s = 10)

    for rank, block in enumerate(significant_blocks):
        x_block, y_block = block[0], block[1]
        clust_x_block = [x_ori2clust_idx[idx] for idx in x_block]
        clust_y_block = [y_ori2clust_idx[idx] for idx in y_block]
        ax.vlines([min(clust_y_block), max(clust_y_block)+1], min(clust_x_block), max(clust_x_block)+1, color='black', linewidths=block_border_width, zorder=3)
        ax.hlines([min(clust_x_block), max(clust_x_block)+1], min(clust_y_block), max(clust_y_block)+1, color='black', linewidths=block_border_width, zorder=3)
        # add rank text
        text_content = str(rank + 1)
        text_size = (min(max(clust_y_block) - min(clust_y_block),
                         max(clust_x_block) - min(clust_x_block)) + 1)*text_scale
        text = ax.text(
            np.mean(clust_y_block) + 0.5, np.mean(clust_x_block) + 0.5,
            text_content, size=text_size, color='white', ha='center', va='center', weight='bold')
        text.set_path_effects([
            path_effects.Stroke(linewidth=3, foreground='black'),
            path_effects.Normal(),
        ])
    plt.savefig(output_file, format=output_file.split('.')[-1].lower(), bbox_inches='tight')

def report_all_associations(dir_name, x_features, y_features, sim_table, pval_table, qval_table, output_file='all_associations.txt'):
    '''Store the association between each feature in X and Y along with p-values and q-values, given:
    - dir_name          : output directory name
    - {x,y}_features    : feature names of {x,y}
    - sim_table         : similarity table with size [len(x_features), len(y_features)]
    - pval_table        : pvalue table with size [len(x_features), len(y_features)]
    - qval_table        : qvalue table with size [len(x_features), len(y_features)]
    - output_file       : the output file name

    Store a .txt file that contains a table with column titles:
        'X_features', 'Y_features', 'association', 'p-values', 'q-values'
    '''
    filepath = join(dir_name, output_file)
    # initiate arrays for generating a pandas DataFrame later
    list_x_features, list_y_features = [], []
    list_association = []
    list_pvals, list_qvals = [], []
    for i in range(len(x_features)):
        for j in range(len(y_features)):
            list_x_features.append(x_features[i])
            list_y_features.append(y_features[j])
            list_association.append(sim_table[i][j])
            list_pvals.append(pval_table[i][j])
            list_qvals.append(qval_table[i][j])
    # create a pandas DataFrame
    df = pd.DataFrame(data={
        'X_features' : list_x_features,
        'Y_features' : list_y_features,
        'association': list_association,
        'p-values'   : list_pvals,
        'q-values'   : list_qvals,
    })
    # store into a file
    df.to_csv(filepath, sep='\t', index=False)

def report_significant_clusters(dir_name, significant_blocks, scores, x_features, y_features,
                                score_label='best_adjusted_pvalue', output_file='sig_clusters.txt'):
    '''Store only the significant clusters, given:
    - dir_name          : output directory name
    - significant_blocks: a list of significant blocks in the original indices, e.g.,
                          [[[2], [0]], [[0,1], [1]]] --> two blocks
    - scores            : a list of scores for each significant blocks
    - {x,y}_features    : feature names of {x,y}
    - output_file       : the output file name
    Store a .txt file that contains a table with column titles:
        'cluster_X', 'cluster_Y'
    '''
    score_label = score_label.replace(' ', '_')
    filepath = join(dir_name, output_file)
    # initiate arrays for generating a pandas DataFrame later
    list_x_clust, list_y_clust = [], []
    for block in significant_blocks:
        list_x_clust.append(';'.join([x_features[idx] for idx in block[0]]))
        list_y_clust.append(';'.join([y_features[idx] for idx in block[1]]))
    # create a pandas DataFrame
    df = pd.DataFrame(data={
        'cluster_rank': [i+1 for i in range(len(significant_blocks))],
        'cluster_X'   : list_x_clust,
        'cluster_Y'   : list_y_clust,
        score_label   : scores
    })
    # store into a file
    df.to_csv(filepath, sep='\t', index=False)

def generate_lattice_plot(x_data, y_data, x_ori_data, y_ori_data, x_features, y_features, x_types, y_types,
                            title, output_file, axis_stretch=1e-5, plot_size=4, n_pairs_to_show=105):
    '''Generate and store lattice plot for each associationn, given:
    - {x,y}_data    : the data for features in {x,y} involved in the association in numpy (the discretized data if discretized)
    - {x,y}_ori_data: the original data for features in {x,y} involved in the association in numpy
    - {x,y}_features: the names of the features in {x,y} involved in the association
    - {x,y}_types   : data types in {x,y}'s features, which determines scatterplot/boxplot/confusion matrix
    - title         : the plot title
    - output_file   : the output file name
    - axis_stretch  : stretch both axes of continuous data by the value, e.g., x --> [x_min - axis_stretch, x_max + axis_stretch]
    - plot_size     : the size of each plot - the figsize would be (# features*plot_size, # features*plot_size)
    '''
    if (len(x_data) != len(x_features) and len(x_data) != len(x_types)) or \
        (len(y_data) != len(y_features) and len(y_data) != len(y_types)):
        raise ValueError('{x,y}_data should have the same length as {x,y}_features and {x,y}_types!')
    row_num = len(x_data) + len(y_data)

    if (row_num**2 - row_num) > n_pairs_to_show:
        n_show = n_pairs_to_show
    else:
        n_show = row_num**2 - row_num
    show_pairs = np.random.choice(a = range(row_num**2 - row_num), size = n_show, replace = False)
    show = -1

    sns.set_style('white')
    fig, axs = plt.subplots(row_num, row_num, figsize=(row_num*plot_size, row_num*plot_size))
    # combine all data
    all_data = np.concatenate((x_data, y_data), axis=0)
    all_ori_data = np.concatenate((x_ori_data, y_ori_data), axis=0)
    all_features = list(x_features) + list(y_features)
    all_types = list(x_types) + list(y_types)
    for i in range(row_num):
        for j in range(row_num):
            if i < j:
                axs[i,j].axis('off')
                continue
            if i == j:
                # 1) plot a histogram if i == j
                if all_types[i] == float:
                    # continuous data: show CDF of the original data
                    # - if discretized, show borders between categories
                    y_min, y_max = 0.0, 1.1
                    x_min, x_max = np.nanmin(all_ori_data[i]) - axis_stretch, np.nanmax(all_ori_data[i]) + axis_stretch
                    sorted_data = np.sort(np.concatenate(([x_min], np.unique(all_ori_data[i]), [x_max])))
                    cdf_line = [(all_ori_data[i] <= val).sum()/len(all_ori_data[i]) for val in sorted_data]
                    sns.lineplot(x=sorted_data, y=cdf_line, ax=axs[i,j], zorder=1)
                    try:
                        np.testing.assert_equal(all_ori_data[i], all_data[i])
                    except: # discretized
                        ori_data, disc_data = np.array(all_ori_data[i]), np.array(all_data[i])
                        border_x = [np.nanmax(ori_data[disc_data ==  x]) for x in range(int(disc_data.min()), int(disc_data.max()))]
                        border_y = [len(ori_data[disc_data <=  x])*1.0/len(ori_data) for x in range(int(disc_data.min()), int(disc_data.max()))]
                        # normalize border_x
                        border_x_norm = [(x - x_min) / (x_max - x_min) for x in border_x]
                        border_y_norm = [(y - y_min) / (y_max - y_min) for y in border_y]
                        for k in range(len(border_x)):
                            axs[i,j].axvline(x=border_x[k], ymax=border_y_norm[k], color='k', ls='--', zorder=2)
                            axs[i,j].axhline(y=border_y[k], xmin=border_x_norm[k], color='k', ls='--', zorder=2)
                    axs[i,j].set_xlim(x_min, x_max)
                    axs[i,j].set_ylim(y_min, y_max)
                else:
                    # categorical data: bins should not be set by default
                    sns.countplot(x=all_data[i], ax=axs[i,j])
            elif all_types[i] == all_types[j]:
                show += 1
                if not show in show_pairs:
                    axs[i,j].axis('off')
                    continue
                if all_types[i] == object:
                    # 2) plot confusion matrix if both are categorical
                    conf_mat = np.zeros((max(all_data[i]).astype(np.int64)+1, max(all_data[j]).astype(np.int64)+1))
                    for k in range(len(all_data[i])):
                        conf_mat[all_data[i,k].astype(np.int64), all_data[j,k].astype(np.int64)] += 1
                    sns.heatmap(conf_mat, ax=axs[i,j], annot=True, cbar=False, cmap='Blues', linewidths=0.1, linecolor='gray')
                else:
                    # 3) plot scatterplot if both are continuous
                    sns.scatterplot(x=all_ori_data[j], y=all_ori_data[i], ax=axs[i,j])
                    axs[i,j].set_xlim(np.nanmin(all_ori_data[j]) - axis_stretch, np.nanmax(all_ori_data[j]) + axis_stretch)
                    axs[i,j].set_ylim(np.nanmin(all_ori_data[i]) - axis_stretch, np.nanmax(all_ori_data[i]) + axis_stretch)
            else:
                # 4) plot boxplot if the data are mixed
                if all_types[j] == float:
                    sns.boxplot(x=all_ori_data[j], y=all_data[i], orient='h', color='w', ax=axs[i,j])
                    sns.stripplot(x=all_ori_data[j], y=all_data[i], jitter=0.25, orient='h', ax=axs[i,j])
                    axs[i,j].set_xlim(np.nanmin(all_ori_data[j]) - axis_stretch, np.nanmax(all_ori_data[j]) + axis_stretch)
                else:
                    sns.boxplot(x=all_data[j], y=all_ori_data[i], color='w', ax=axs[i,j])
                    sns.stripplot(x=all_data[j], y=all_ori_data[i], jitter=0.25, ax=axs[i,j])
                    axs[i,j].set_ylim(np.nanmin(all_ori_data[i]) - axis_stretch, np.nanmax(all_ori_data[i]) + axis_stretch)
            # add y-ticks on the right side on histogram plots
            if i == j:
                axs[i,j].yaxis.tick_right()
                axs[i,j].yaxis.set_label_position('right')
                if i == 0:
                    axs[i,j].set_ylabel('\n\nCumulative probability')
                else:
                    axs[i,j].set_ylabel('Cumulative probability')
                if i != row_num - 1: axs[i,j].set_xticks([])
            else:
                # remove ticks from inner plots
                if j != 0: axs[i,j].set_yticks([])
                if i != row_num - 1: axs[i,j].set_xticks([])
            # add labels to outer  plots
            if j == 0:
                if i != 0:
                    ax2 = axs[i,j].twinx()
                    ax2.set_yticks([])
                    ax2.yaxis.set_label_position('left')
                    ax2.set_ylabel(all_features[i], fontdict=dict(weight='bold'))
            if i == row_num - 1: axs[i,j].set_xlabel(all_features[j], fontdict=dict(weight='bold'))
            for spine in axs[i,j].spines:
                axs[i,j].spines[spine].set_visible(True)
    # align labels
    fig.align_xlabels()
    fig.align_ylabels()
    fig.suptitle(title)
    plt.subplots_adjust(wspace=.05, hspace=.05)
    plt.savefig(output_file)
    plt.close()



import numpy as np
import sys
import itertools
import functools

def is_densely_associated(block, fnr_thresh=0.1):
    '''Check if the significance block (boolean array) is densely associated, given the
       false negative rate threshold
    '''
    # ensure that block is a numpy boolean array
    if type(block) is not np.ndarray or block.dtype != 'bool':
        raise ValueError('block argument should be a boolean Numpy array!')
    return(np.sum(block) / block.size >= 1 - fnr_thresh)

def calc_gini_impurity(ar):
    '''Calculate gini impurity: 1 - proportion(True)^2 - proportion(False)^2
    Arg:
    - ar: a binary array (can be any dimension)
    '''
    # ensure that ar is a boolean array
    ar = np.asarray(ar)
    if ar.dtype != 'bool':
        raise ValueError('ar argument should be a boolean array!')
    prop_true  = np.sum(ar) / ar.size
    prop_false = 1 - prop_true
    return(1 - prop_true**2 - prop_false**2)

def calc_weighted_gini_impurity(lists):
    '''Calculate weighted gini impurity given an array of lists
    Arg:
    - lists: a list of binary arrays
    '''
    total_size = sum([len(list_i) for list_i in lists])
    weighted_scores = [len(list_i)*1.0/total_size * calc_gini_impurity(list_i) for list_i in lists]
    return(sum(weighted_scores))

def bifurcate(tree):
    '''Bifurcate a tree to its two branches
    - Arg: tree (ClusterNode object)
    Returns an array with length <=2 containing ClusterNode branches
      If tree is a leaf, it should return an empty array
    '''
    return([branch for branch in [tree.get_left(), tree.get_right()] if branch is not None])

def bifurcate_one(x_tree, y_tree, fdr_reject_table, splitting_diagnostic_mode=False, gini_uncertainty_level = 0.01):
    '''Given two trees, bifurcate only one:
    - 1) if both are leaves, return (None, None)
    - 2) if one of them is a leaf, return the bifurcated non-leaf tree
    - 3) if both are non-leaves, bifurcate the one that produces lower gini impurity
    Args:
    - x_tree, y_tree   = ClusterNode objects
    '''
    x_branches = bifurcate(x_tree)
    y_branches = bifurcate(y_tree)
    # 1) if both are leaves
    if len(x_branches) == 0 and len(y_branches) == 0: return(None, None)
    # 2) if one of them is a leaf
    if len(x_branches) == 0:
        return([x_tree], y_branches)
    if len(y_branches) == 0:
        return(x_branches, [y_tree])
    # 3) if both are non-leaves
    x_features = x_tree.pre_order()
    y_features = y_tree.pre_order()
    # split1 = split x_tree
    split1_blocks = [[fdr_reject_table[feature_x, feature_y] for feature_x, feature_y in itertools.product(branch.pre_order(), y_features)]
                                                             for branch in x_branches]
    split1_gini = calc_weighted_gini_impurity(split1_blocks)
    # split2 = split y_tree
    split2_blocks = [[fdr_reject_table[feature_x, feature_y] for feature_x, feature_y in itertools.product(x_features, branch.pre_order())]
                                                             for branch in y_branches]
    split2_gini = calc_weighted_gini_impurity(split2_blocks)

    # If the gini impurity of both proposed splits are close, split the larger branch
    # It would be better to make the threshold sensitive to the size and proportions of the proposed splits, but it's hard to get an unbiased interval.
    if gini_uncertainty_level > 0 and abs(split1_gini - split2_gini) < gini_uncertainty_level and (x_tree.get_count() * y_tree.get_count()) > 25:
        return((x_branches, [y_tree]) if x_tree.get_count() > y_tree.get_count() else ([x_tree], y_branches))

    if splitting_diagnostic_mode:
        print(split1_gini, file = sys.stderr)
        print(split2_gini, file = sys.stderr)
        print("\n", file = sys.stderr)

    return((x_branches, [y_tree]) if split1_gini < split2_gini else ([x_tree], y_branches)) 

def compare_and_find_dense_block(X, Y, fdr_reject_table, fnr_thresh=0.1, splitting_diagnostic_mode=False, gini_uncertainty_level = .02):
    '''Given another HierarchicalTree object Y, compare and find
    densely-associated block from the top of the hierarchy;

    Densely-associated block = (1 - FNR)% of pairwise association are
      FDR significant
    Args:
    - X               : X hierarchical tree (ClusterNode object)
    - Y               : Y hierarchical tree (ClusterNode object)
    - fdr_reject_table: a boolean table where True = reject H0
    - fnr_thresh      : false negative rate threshold
    '''

    def _check_iter_block(x_tree, y_tree, gini_uncertainty_level, splitting_diagnostic_mode=False):
        '''Check block iteratively until:
        - a densely-associated block is found
        - x_tree and y_tree are leaves
        Append densely-associated blocks to final_blocks array

        Terminate when:
        1) report block
        2) can no longer bifurcate
        '''
        X_features = x_tree.pre_order()
        Y_features = y_tree.pre_order()
        block_fdr_reject = fdr_reject_table[X_features,:][:,Y_features]

        if splitting_diagnostic_mode:
            print(X_features, file = sys.stderr)
            print(Y_features, file = sys.stderr)
            print("\n", file = sys.stderr)
        if is_densely_associated(block_fdr_reject, fnr_thresh):
            # 1) terminate when the block is reported
            final_blocks.append([X_features, Y_features])
            return
        x_branches, y_branches = bifurcate_one(x_tree, y_tree, fdr_reject_table, splitting_diagnostic_mode, gini_uncertainty_level=gini_uncertainty_level)
        if x_branches is None and y_branches is None:
            # 2) terminate when both trees can no longer bifurcate
            return
        for x_branch, y_branch in itertools.product(x_branches, y_branches):
            _check_iter_block(x_branch, y_branch, gini_uncertainty_level, splitting_diagnostic_mode)

    final_blocks = []
    _check_iter_block(X, Y, gini_uncertainty_level, splitting_diagnostic_mode)
    return(final_blocks)

def trim_block(block, fdr_reject_table):
    '''Trim the sides if all points on the side are insignificant
    # TODO: set a threshold?
    '''
    x_features, y_features = np.array(block[0]), np.array(block[1])
    x_start, x_end = 0, len(x_features)-1
    y_start, y_end = 0, len(y_features)-1
    while x_start < x_end:
        if fdr_reject_table[x_features[x_start],:][y_features].sum() > 0: break
        x_start += 1
    while x_end > x_start:
        if fdr_reject_table[x_features[x_end],:][y_features].sum() > 0: break
        x_end -= 1
    while y_start < y_end:
        if fdr_reject_table[:, y_features[y_start]][x_features].sum() > 0: break
        y_start += 1
    while y_end > y_start:
        if fdr_reject_table[:, y_features[y_end]][x_features].sum() > 0: break
        y_end -= 1
    return([[x_features[x] for x in range(x_start, x_end+1)], [y_features[y] for y in range(y_start, y_end+1)]])


from os.path import join, isdir
import os
import sys

import shutil

def create_dir(dir_name, verbose=True):
    '''Creating a new directory; if the directory already exists, just let it be~
    '''
    if isdir(dir_name): return
    try:
        if verbose: print('Creating a new directory %s' % dir_name, file = sys.stderr)
        os.mkdir(dir_name)
    except EnvironmentError:
        sys.exit('Unable to create directory %s' % dir_name)

def reset_dir(dir_name, verbose=True):
    '''Remove any existing directory with the same name and create a new one
    '''
    if isdir(dir_name):
        try:
            if verbose: print('Directory %s exists; deleting...' % dir_name, file = sys.stderr)
            shutil.rmtree(dir_name)
        except EnvironmentError:
            sys.exit('Unable to remove directory %s' % dir_name)
    # create a new directory
    create_dir(dir_name, verbose)


# load configurations from config.yaml
import yaml
from os.path import dirname, abspath, join
import sys

yaml_file = join(dirname(abspath(__file__)), 'config.yaml')

class Struct:
    def __init__(self, **entries):
        self.__dict__.update(entries)

with open(yaml_file, 'r') as stream:
    try:
        config = yaml.safe_load(stream)
    except yaml.YAMLError as err:
        print(err, file = sys.stderr)

def update_config(attribute, **args):
    vals = getattr(config, attribute)
    for key in args:
        if key not in vals:
            raise KeyError('%s not found in config.%s' % (key, attribute))
        vals[key] = args[key]
    setattr(config, attribute, vals)

config = Struct(**config)


'''Generate hallagram on console environment
'''

import argparse
import sys
import numpy as np
from os.path import join, splitext
import pkg_resources

from halla.utils.report import generate_hallagram, generate_clustermap
from .loader import HAllAPartialLoader

def parse_argument(args):
    parser = argparse.ArgumentParser(
        description='HAllA clustermap/hallagram generator')

    # --load parameters--
    parser.add_argument(
        '-i', '--input',
        help='Path to the output directory',
        required=True)
    parser.add_argument(
        '-c', '--clustermap',
        help='Generate a clustermap instead of a hallagram',
        action='store_true', required=False)
    parser.add_argument(
        '--x_dataset_label',
        help='Label for X dataset',
        default='', required=False)
    parser.add_argument(
        '--y_dataset_label',
        help='Label for Y dataset',
        default='', required=False)
    parser.add_argument(
        '--cbar_label',
        help='Label for the colorbar',
        default='', required=False)
    parser.add_argument(
        '--cmap',
        help='Colormap',
        default='RdBu_r', required=False)
    parser.add_argument(
        '-n', '--block_num',
        help='Number of top clusters to show (for hallagram only); if -1, show all clusters',
        default=50, type=int, required=False)
    parser.add_argument(
        '--mask',
        help='Mask the hallagram/clustermap',
        action='store_true', required=False)
    parser.add_argument(
        '--trim',
        help='Trim hallagram to features containing at least one significant block',
        default=True,
        type=bool, required=False)
    parser.add_argument(
        '--text_scale',
        help='Significant cluster text scale',
        default=10, type=float, required=False)
    parser.add_argument(
        '--block_border_width',
        help='Significant cluster border width',
        default=1.5, type=float, required=False)
    parser.add_argument(
        '-o', '--output',
        help='Path to output file under the HAllA/AllA result directory; default: hallagram or clustermap',
        default='', required=False)
    parser.add_argument(
        '--fdr_alpha',
        help='FDR threshold',
        default=0.05, type=float, required=False)
    parser.add_argument(
        '--show_lower',
        help='Show multi-member blocks ranked below block_num as grey outlined boxes',
        default=True, type=bool, required=False)
    parser.add_argument(
        '--force_x_ft',
        nargs = '+',
        help='Force specific x features to be included, irrespective of trim setting.',
        default=None,
        required=False)
    parser.add_argument(
        '--force_y_ft',
        nargs = '+',
        help='Force specific y features to be included, irrespective of trim setting.',
        default=None,
        required=False)
    parser.add_argument(
        '--dpi',
        help='Figure DPI',
        default=100, type=float, required=False)
    parser.add_argument(
        '--suppress_numbers',
        help='If True, don\'t number the blocks',
        action='store_true', required=False,
        default=False)
    parser.add_argument(
        "--plot_file_type",
        help = "File type of hallagram output. Ignored if specified by --output",
        default = "pdf",
        dest="plot_type",
        required = False
    )

    params = parser.parse_args()
    if params.block_num == -1: params.block_num = None
    if params.output == '':
        params.output = 'clustermap.pdf' if params.clustermap else 'hallagram'
    return(params)

def main():
    params = parse_argument(sys.argv)
    input_dir = params.input
    output_file = join(input_dir, params.output)
    loader = HAllAPartialLoader(input_dir)
    if params.clustermap:
        if loader.name == 'AllA':
            raise ValueError('Cannot generate a clustermap from AllA result.')
        if max(loader.sim_table.shape) > 500:
            raise ValueError('The dimension is too large - please generate a hallagram instead.')
        generate_clustermap(loader.significant_blocks,
                            loader.X_features,
                            loader.Y_features,
                            loader.X_linkage,
                            loader.Y_linkage,
                            loader.sim_table,
                            x_dataset_label=params.x_dataset_label,
                            y_dataset_label=params.y_dataset_label,
                            fdr_reject_table = loader.fdr_reject_table,
                            cbar_label=params.cbar_label,
                            text_scale=params.text_scale,
                            block_border_width=params.block_border_width,
                            mask=params.mask,
                            cmap=params.cmap,
                            output_file=output_file)
    else:
        block_num = params.block_num
        if block_num is None:
            block_num = len(loader.significant_blocks)
        else:
            block_num = min(block_num, len(loader.significant_blocks))
        if block_num > 500:
            raise ValueError('The number of blocks to show is too large, please input block # < 300')
        out_split = splitext(output_file)
        if out_split[1] == "":
            pt = params.plot_type
        else:
            pt = out_split[1]

        generate_hallagram(loader.significant_blocks,
                           loader.X_features,
                           loader.Y_features,
                           loader.X_tree.pre_order() if loader.name == 'HAllA' else [idx for idx in range(loader.X.shape[0])],
                           loader.Y_tree.pre_order() if loader.name == 'HAllA' else [idx for idx in range(loader.Y.shape[0])],
                           loader.sim_table,
                           fdr_reject_table = loader.fdr_reject_table,
                           block_num = params.block_num,
                           x_dataset_label=params.x_dataset_label,
                           y_dataset_label=params.y_dataset_label,
                           cbar_label=params.cbar_label,
                           text_scale=params.text_scale,
                           block_border_width=params.block_border_width,
                           mask=params.mask,
                           plot_type=out_split[1],
                           trim=params.trim,
                           force_x_ft=params.force_x_ft,
                           force_y_ft=params.force_y_ft,
                           suppress_numbers=params.suppress_numbers,
                           dpi=params.dpi,
                           cmap=params.cmap,
                           output_file=out_split[0])

if __name__ == "__main__":
    main()



from sklearn.metrics import normalized_mutual_info_score, mutual_info_score
from scipy.stats import pearsonr, spearmanr, chi2_contingency
from scipy.spatial.distance import pdist, squareform
from pandas import crosstab
import numpy as np

import rpy2.robjects as robjects
from rpy2.robjects.packages import importr
from rpy2.robjects.vectors import FloatVector

XICOR = importr('XICOR')

def remove_missing_values(x, y):
    '''Given x and y all in numpy arrays, remove pairs that contain missing values
    '''
    # nan != nan = TRUE
    nas = np.logical_or(x != x, y != y)
    return(x[~nas], y[~nas])

'''Similarity wrapper functions (note: not distance!)

Given x, y, returns a tuple (similarity, p-value);
p-value will be None if not provided
'''
def nmi(x, y, return_pval=False):
    '''normalized mutual information, ranging from [0 .. 1]
    0: no mutual information; 1: perfect correlation
    '''
    x, y = remove_missing_values(x, y)
    if (np.unique(x).shape[0] == 1 or np.unique(y).shape[0] == 1):
        if return_pval: return(0,1)
        return(0)
    if return_pval:
        ch2, p, dof, ex = chi2_contingency(crosstab(x,y))
        return(normalized_mutual_info_score(x, y), p)
    return(normalized_mutual_info_score(x, y))

def mutual_info(x,y, return_pval=False):
    x, y = remove_missing_values(x, y)
    if (np.unique(x).shape[0] == 1 or np.unique(y).shape[0] == 1):
        if return_pval: return(0,1)
        return(0)

    if return_pval:
        ch2, p, dof, ex = chi2_contingency(crosstab(x,y))
        return(mutual_info_score(x, y), p)
    return(mutual_info_score(x, y))

def pearson(x, y, return_pval=False):
    x, y = remove_missing_values(x, y)
    if (np.unique(x).shape[0] == 1 or np.unique(y).shape[0] == 1):
        if return_pval: return(0,1)
        return(0)
    corr, pval = pearsonr(x, y)
    # TODO: enable tuning whether correlation should always be positive or not
    if return_pval: return(corr, pval)
    return(corr)

def spearman(x, y, return_pval=False):
    x, y = remove_missing_values(x, y)
    if (np.unique(x).shape[0] == 1 or np.unique(y).shape[0] == 1):
        if return_pval: return(0,1)
        return(0)
    corr, pval = spearmanr(x, y)
    # TODO: enable tuning whether correlation should always be positive or not
    if return_pval: return(corr, pval)
    return(corr)

def distcorr(x, y, return_pval=False):
    '''Perform distance correlation [0 .. 1]
    def src: https://en.wikipedia.org/wiki/Distance_correlation
    distance corr = 0 iif x and y are independent
    distance corr = 1 implies that dimensions of the linear subspaces spanned by x & y respectively are
      almost surely equal and if we assume that these subspaces are equal, then in this subspace
      y = A + bCx for some vector A, scalar b, and orthonormal matrix C
    code src: https://gist.github.com/satra/aa3d19a12b74e9ab7941 - much faster than the library dcor
    '''
    x, y = remove_missing_values(x, y)
    if (np.unique(x).shape[0] == 1 or np.unique(y).shape[0] == 1):
        if return_pval: return(0,1)
        return(0)
    x, y = np.atleast_1d(x), np.atleast_1d(y)
    # if 1D - add dummy axis
    if np.prod(x.shape) == len(x): x = x[:, None]
    if np.prod(y.shape) == len(y): y = y[:, None]
    x = np.atleast_2d(x)
    y = np.atleast_2d(y)
    n = x.shape[0]
    if x.shape[0] != y.shape[0]:
        raise ValueError('Number of samples must match')
    a, b = squareform(pdist(x)), squareform(pdist(y))
    A = a - a.mean(axis=0)[None, :] - a.mean(axis=1)[:, None] + a.mean()
    B = b - b.mean(axis=0)[None, :] - b.mean(axis=1)[:, None] + b.mean()

    dcov2_xy = (A * B).sum()/float(n * n)
    dcov2_xx = (A * A).sum()/float(n * n)
    dcov2_yy = (B * B).sum()/float(n * n)
    dcor = np.sqrt(dcov2_xy)/np.sqrt(np.sqrt(dcov2_xx) * np.sqrt(dcov2_yy))

    if return_pval: return(dcor, None)
    return(dcor)

def xicor(x,y,return_pval=False):
    x, y = remove_missing_values(x, y)
    if return_pval: xi, sd, pval = XICOR.xicor(FloatVector(x),FloatVector(y), pvalue = True)
    else: xi = XICOR.xicor(FloatVector(x),FloatVector(y), pvalue = False)
    if return_pval: return(xi[0],pval[0])
    return(xi[0])

def symmetric_xicor(x,y):
    '''
    This is used for computing trees. It isn't used for testing hypotheses
    because there's no closed form expression for the p-value.
    '''
    xi_xy = xicor(x,y)
    xi_yx = xicor(y,x)
    return(max(xi_xy, xi_yx))

'''Constants
'''
SIM_FUNCS = {
    'nmi': nmi,
    'mi': mutual_info,
    'pearson': pearson,
    'spearman': spearman,
    'dcor': distcorr,
    'xicor': xicor,
    'symmetric_xicor': symmetric_xicor
}

PVAL_PROVIDED = {
    'nmi': True,
    'mi': True,
    'pearson': True,
    'spearman': True,
    'dcor': False,
    'xicor': True,
}

def get_similarity_function(metric):
    '''Retrieve the right distance function
    according to the metric
    '''
    metric = metric.lower()
    if metric not in SIM_FUNCS:
        raise KeyError('The similarity metric is not available...')
    # only return the similarity scores
    return(SIM_FUNCS[metric])

def does_return_pval(metric):
    return(PVAL_PROVIDED[metric])

def similarity2distance(scores, set_abs=True, convert_func=None):
    '''Convert similarity scores to distance; by default, the conversion formula would be:
       dist = 1 - transform(similarity scores) where transform is abs(scores) if set_abs is True else scores
       reference of converting correlation by 1 - scores: https://www.ncbi.nlm.nih.gov/pmc/articles/PMC4498680/
    - scores      : the similarity values to be converted in a numpy array
    - set_abs     : if True, set scores to absolute
    - convert_func: if specified, use the provided function to convert instead
    '''
    if type(scores) is not np.ndarray:
        raise ValueError('scores argument should be a numpy array!')
    if convert_func is not None:
        return(convert_func(scores))
    return(1 - (np.abs(scores) if set_abs else scores))



import pkg_resources  # part of setuptools to retrieve version
import datetime
import sys
from os.path import join

class HAllALogger(object):
    def __init__(self, name, config):
        self.name = name
        self.verbose = config.output['verbose']
        self.durations = [] # details on durations for certain steps
        self.results = []   # details on results from certain steps
        self.log_config(config)

    def log_config(self, config, return_text=False):
        '''Log configuration parameters by iterating through config (Struct) variable
        if return_text is True, return text instead of printing anything
        '''
        def format_text(text):
            return(text.replace('_', ' '))
        log_txt = ''
        config_dict = config.__dict__
        for key_1 in config_dict:
            log_txt += '  %s:\n' % format_text(key_1)
            for key_2 in config_dict[key_1]:
                log_txt += '    %-*s: ' % (30, format_text(key_2)) + str(config_dict[key_1][key_2]) + '\n'
        if return_text:
            return(log_txt)
        if not self.verbose: return
        print('Setting config parameters (irrelevant parameters will be ignored)...', file = sys.stderr)
        print(log_txt, file = sys.stderr)

    def log_step_start(self, message, sub=False):
        '''Log the beginning of a step (or a sub-step):
        - a (main) step: == message ==
        - a (sub)  step: -- message --
        '''
        decorator = '--' if sub else '=='
        if self.verbose: print('%s %s %s' % (decorator, message, decorator), file = sys.stderr)

    def log_step_end(self, label, dur_second, sub=False):
        '''Log the end of a step (or a sub-step):
        - a (main) step: == Completed; total duration: time ==
        - a (sub)  step: -- Completed; total duration: time --
        The (label, dur_second) is added to self.durations array to be printed in performance.txt
        '''
        decorator = '--' if sub else '=='
        dur_str = str(datetime.timedelta(seconds=dur_second))
        self.durations.append((label, dur_second))
        if self.verbose:
            print('%s Completed; total duration: %s %s\n' % (decorator, dur_str, decorator), file = sys.stderr)

    def log_message(self, message):
        '''Log a message
        '''
        if self.verbose: print(message, file = sys.stderr)

    def log_result(self, label, content):
        '''Log results with prefix '  '
        The (label, content) is added to self.results array to be printed in performance.txt
        '''
        self.results.append((label, content))
        if self.verbose: print('* ', label, content, file = sys.stderr)

    def write_performance_log(self, dir_name, config, file_name='performance.txt'):
        '''Write performance.txt which contains:
        - halla version
        - configuration parameters
        - results details
        - durations details
        '''
        performance_txt = 'HAllA version:\t' + pkg_resources.require('HAllA')[0].version + '\n'
        # add config details
        performance_txt += '\n--Configuration parameters--\n' + self.log_config(config, return_text=True)
        # add result details
        performance_txt += '\n--Results--\n' + \
                           '\n'.join(['%-*s: ' % (60, item[0]) + str(item[1]) for item in self.results]) + '\n'
        # add duration details
        performance_txt += '\n--Durations--\n' + \
                           '\n'.join(['%-*s: ' % (60, item[0]) + str(datetime.timedelta(seconds=item[1])) \
                                      for item in self.durations]) + '\n'
        tot_dur = sum(item[1] for item in self.durations)
        performance_txt += '%-*s: ' % (60, 'Total execution time') + str(datetime.timedelta(seconds=tot_dur))
        file = open(join(dir_name, file_name), 'w')
        file.write(performance_txt)
        file.close()



import unittest
import numpy as np
import pandas as pd
from scipy.stats import zscore
from halla.utils import data

from utils import compare_numpy_array

class TestDataUtils(unittest.TestCase):

    '''Tests on eval_type functions
    '''
    def test_eval_type_input_df(self):
        '''Tests if it throws an error when non-pd.DataFrame arg is given
        '''
        self.assertRaises(ValueError, data.eval_type, np.zeros((5, 5)))
    
    def test_eval_type_result_1(self):
        df = pd.DataFrame(data={ 'sample_1': ['24', 'a'],
                                 'sample_2': ['13.6', 'b'],
                                 'sample_3': [np.nan, 'a'] })
        _, types = data.eval_type(df)
        self.assertTrue(compare_numpy_array(types, np.array([float, object])))
    
    def test_eval_type_result_2(self):
        df = pd.DataFrame(data={ 'sample_1': ['24', 'a'],
                                 'sample_2': ['13.6', 'b'],
                                 'sample_3': [np.nan, 'a'] })
        updated_df, _ = data.eval_type(df)
        self.assertTrue(compare_numpy_array(updated_df.iloc[0].to_numpy(), np.array([24, 13.6, np.nan])))
    
    def test_eval_type_result_3(self):
        df = pd.DataFrame(data={ 'sample_1': ['24', 'a'],
                                 'sample_2': ['13.6', 'b'],
                                 'sample_3': [np.nan, 'a'] })
        updated_df, _ = data.eval_type(df)
        self.assertTrue(compare_numpy_array(updated_df.iloc[1].to_numpy(), np.array(['a', 'b', 'a'])))

    '''Tests on discretization
    '''
    def test_discretize_vector_categorical(self):
        ar = np.array(['a', 'c', 'b', 'b', 'c', 'b', 'a', 'd', 'a', 'd', 'e'])
        res = data.discretize_vector(ar, ar_type=object)
        expected_res = np.array([0, 1, 2, 2, 1, 2, 0, 3, 0, 3, 4])
        self.assertTrue(compare_numpy_array(res, expected_res))
    
    def test_discretize_vector_categorical_missing_data(self):
        ar = np.array(['a', 'c', 'b', 'b', 'c', np.nan, 'a', 'd', 'a', 'd', 'e', np.nan])
        res = data.discretize_vector(ar, ar_type=object)
        expected_res = np.array([0, 1, 2, 2, 1, 3, 0, 4, 0, 4, 5, 3])
        self.assertTrue(compare_numpy_array(res, expected_res))

    def test_dicretize_vector_continuous_quantile(self):
        ar = np.array([1, 1, 1, 1, 1, 1, 2, 2, 2, 3, 3, 3, np.nan, 4, 4, 4, 4, 5, 5, 5, 6, 6, np.nan])
        res = data.discretize_vector(ar, func='quantile', num_bins=3)
        expected_res = np.array([0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 3])
        self.assertTrue(compare_numpy_array(res, expected_res))

    def test_dicretize_vector_continuous_jenks(self):
        ar = np.array([1, 1, 1, 1, 1, 1, 2, 2, 2, 3, 3, 3, np.nan, 4, 4, 4, 4, 5, 5, 5, 6, 6, np.nan])
        res = data.discretize_vector(ar, func='jenks', num_bins=3)
        expected_res = np.array([0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 3])
        self.assertTrue(compare_numpy_array(res, expected_res))
    
    '''Tests on keep_feature function
    '''
    def test_keep_feature_1(self):
        ar = [1, 1, 0, 0, 0, 0, 0, 1, 0, 1]
        df_row = pd.DataFrame(ar).T.iloc[0]
        self.assertFalse(data.keep_feature(df_row, 0.5))
    
    def test_keep_feature_2(self):
        ar = [1, 1, 0, 0, 0, 0, 0, 1, 0, 1]
        df_row = pd.DataFrame(ar).T.iloc[0]
        self.assertFalse(data.keep_feature(df_row, 0.6))
    
    def test_keep_feature_3(self):
        ar = [1, 1, 0, 0, 0, 0, 0, 1, 0, 1]
        df_row = pd.DataFrame(ar).T.iloc[0]
        self.assertTrue(data.keep_feature(df_row, 0.61))
    
    def test_keep_feature_4(self):
        ar = [1, 1, 1, 1, 1, 1, 1, 1]
        df_row = pd.DataFrame(ar).T.iloc[0]
        self.assertTrue(data.keep_feature(df_row, None))
    
    def test_keep_feature_5(self):
        ar = [1, 1, 1, 1, 1, 1, 1, 1]
        df_row = pd.DataFrame(ar).T.iloc[0]
        self.assertFalse(data.keep_feature(df_row, 1))
    
    '''Tests on transform function
    '''
    def test_transform_invalid_func(self):
        '''Tests if it throws an error when an invalid function is given
        '''
        self.assertRaises(ValueError, data.transform, None, None, 'strange_func')

    def test_transform_log_1(self):
        df = pd.DataFrame(data={ 'sample_1': [24, 'a'],
                                 'sample_2': [13.6, 'b'],
                                 'sample_3': [np.nan, 'a'] })
        types = np.array([float, object])
        updated_df = data.transform(df, types, ['log'])
        self.assertTrue(compare_numpy_array(updated_df.iloc[0].to_numpy(), np.log([24, 13.6, np.nan])))
    
    def test_transform_log_2(self):
        df = pd.DataFrame(data={ 'sample_1': [24, 'a'],
                                 'sample_2': [13.6, 'b'],
                                 'sample_3': [np.nan, 'a'] })
        types = np.array([float, object])
        updated_df = data.transform(df, types, ['log'])
        self.assertTrue(compare_numpy_array(updated_df.iloc[1].to_numpy(), np.array(['a', 'b', 'a'])))
    
    def test_transform_log_3(self):
        df = pd.DataFrame(data={ 'sample_1': [24, 'a'],
                                 'sample_2': [-13.6, 'b'],
                                 'sample_3': [np.nan, 'a'] })
        types = np.array([float, object])
        self.assertRaises(ValueError, data.transform, df, types, ['log'])
    
    def test_transform_sqrt(self):
        df = pd.DataFrame(data={ 'sample_1': [24, 'a'],
                                 'sample_2': [-13.6, 'b'],
                                 'sample_3': [np.nan, 'a'] })
        types = np.array([float, object])
        row = np.array([24, -13.6, np.nan])
        updated_df = data.transform(df, types, ['sqrt'])
        self.assertTrue(compare_numpy_array(updated_df.iloc[0].to_numpy(), np.sqrt(np.abs(row)) * np.sign(row)))
    
    def test_transform_multiple_functions(self):
        df = pd.DataFrame(data={ 'sample_1': [24, 'a'],
                                 'sample_2': [-13.6, 'b'],
                                 'sample_3': [np.nan, 'a'],
                                 'sample_4': [10, 'a'] })
        types = np.array([float, object])
        updated_df = data.transform(df, types, ['zscore', 'sqrt'])
        expected_res = zscore(np.array([24, -13.6, np.nan, 10]), nan_policy='omit')
        expected_res = np.sqrt(np.abs(expected_res)) * np.sign(expected_res)
        self.assertTrue(compare_numpy_array(updated_df.iloc[0].to_numpy(), expected_res))
    
    


'''HAllA setup

To install: python setup.py install
'''

import sys

try:
    import setuptools
    from setuptools.command.install import install
except ImportError:
    sys.exit('Please install setuptools.')

VERSION = '0.8.30'
AUTHOR  = 'HAllA Development Team'
MAINTAINER_EMAIL = 'halla-users@googlegroups.com'

class PostInstallCommand(install):
    '''Post-installation for installation mode'''
    def run(self):
        install.run(self)
        # post-install script
        from rpy2.robjects.packages import importr
        try:
            eva = importr('eva')
        except:
            utils = importr('utils')
            utils.chooseCRANmirror(ind=1)
            utils.install_packages('EnvStats')
            utils.install_packages('https://cran.r-project.org/src/contrib/Archive/eva/eva_0.2.5.tar.gz')
            # check if eva has been successfully installed
            eva = importr('eva')
        try:
            XICOR = importr('XICOR')
        except:
            utils = importr('utils')
            utils.chooseCRANmirror(ind=1)
            utils.install_packages("XICOR")
            XICOR = importr('XICOR')

# Installing requirements.txt dependencies
dependencies=[]
requirements = open('requirements.txt', 'r')
for dependency in requirements:
    dependencies.append(str(dependency))
    
setuptools.setup(
    name='HAllA',
    author=AUTHOR,
    author_email=MAINTAINER_EMAIL,
    version=VERSION,
    license='MIT',
    description='HAllA: Hierarchical All-against All Association Testing',
    long_description="Given two high-dimensional 'omics datasets X and Y (continuous and/or categorical features) from the same n biosamples, HAllA (Hierarchical All-against-All Association Testing) discovers densely-associated blocks of features in the X vs. Y association matrix where: 1) each block is defined as all associations between features in a subtree of X hierarchy and features in a subtree of Y hierarchy and 2) a block is densely associated if (1 - FNR)% of pairwise associations are FDR significant (FNR is the pre-defined expected false negative rate)",
    url='https://github.com/biobakery/halla',
    keywords=['halla', 'association testing'],
    platforms=['Linux','MacOS'],
    install_requires=dependencies,
    classifiers=[
        'Programming Language :: Python',
        'Operating System :: MacOS',
        'Operating System :: Unix',
        'Programming Language :: Python :: 3.7',
        'Topic :: Scientific/Engineering :: Bio-Informatics'
      ],
    packages=setuptools.find_packages(),
    package_data={
        'halla': ['config.yaml']
    },
    entry_points={
        'console_scripts': [
            'halla = scripts.halla:main',
            'halladata = scripts.synthetic_data:main',
            'hallagram = scripts.hallagram:main',
            'hallagnostic = scripts.diagnostic_plot:main',
        ]
    },
    cmdclass={
        'install': PostInstallCommand,
    },
    test_suite= 'tests',
 )
